
# distutils: libraries = STD_LIBS
"""
A refine-by-two AMR-specific octree



"""



import numpy as np

cimport cython

# Double up here for def'd functions
cimport numpy as np
cimport numpy as cnp
from libc.stdlib cimport abs, free, malloc

from yt.utilities.lib.fp_utils cimport fclip, fmax, fmin, iclip, imax, imin

import sys
import time


cdef extern from "platform_dep.h":
    # NOTE that size_t might not be int
    void *alloca(int)

cdef extern from "math.h":
    double sqrt(double x)

cdef inline np.float64_t f64max(np.float64_t f0, np.float64_t f1):
    if f0 > f1: return f0
    return f1

cdef struct OctreeNode:
    np.float64_t *val
    np.float64_t weight_val
    np.int64_t pos[3]
    int level
    int nvals
    int max_level # The maximum level under this node with mass.
    OctreeNode *children[2][2][2]
    OctreeNode *parent
    OctreeNode *next
    OctreeNode *up_next

cdef void OTN_add_value(OctreeNode *self,
        np.float64_t *val, np.float64_t weight_val, int level, int treecode):
    cdef int i
    for i in range(self.nvals):
        self.val[i] += val[i]
    self.weight_val += weight_val
    if treecode and val[0] > 0.:
        self.max_level = imax(self.max_level, level)

cdef void OTN_refine(OctreeNode *self, int incremental = 0):
    cdef int i, j, k
    cdef np.int64_t npos[3]
    for i in range(2):
        npos[0] = self.pos[0] * 2 + i
        for j in range(2):
            npos[1] = self.pos[1] * 2 + j
            # We have to be careful with allocation...
            for k in range(2):
                npos[2] = self.pos[2] * 2 + k
                self.children[i][j][k] = OTN_initialize(
                            npos,
                            self.nvals, self.val, self.weight_val,
                            self.level + 1, self, incremental)
    if incremental: return
    for i in range(self.nvals): self.val[i] = 0.0
    self.weight_val = 0.0

cdef OctreeNode *OTN_initialize(np.int64_t pos[3], int nvals,
                        np.float64_t *val, np.float64_t weight_val,
                        int level, OctreeNode *parent, int incremental = 0):
    cdef OctreeNode *node
    cdef int i, j, k
    node = <OctreeNode *> malloc(sizeof(OctreeNode))
    node.pos[0] = pos[0]
    node.pos[1] = pos[1]
    node.pos[2] = pos[2]
    node.nvals = nvals
    node.parent = parent
    node.next = NULL
    node.up_next = NULL
    node.max_level = 0
    node.val = <np.float64_t *> malloc(
                nvals * sizeof(np.float64_t))
    if incremental:
        for i in range(nvals):
            node.val[i] = 0.
        node.weight_val = 0.
    else:
        for i in range(nvals):
            node.val[i] = val[i]
        node.weight_val = weight_val
    for i in range(2):
        for j in range(2):
            for k in range(2):
                node.children[i][j][k] = NULL
    node.level = level
    return node

cdef void OTN_free(OctreeNode *node):
    cdef int i, j, k
    for i in range(2):
        for j in range(2):
            for k in range(2):
                if node.children[i][j][k] == NULL: continue
                OTN_free(node.children[i][j][k])
    free(node.val)
    free(node)

cdef class Octree:
    cdef int nvals
    cdef np.int64_t po2[80]
    cdef OctreeNode ****root_nodes
    cdef np.int64_t top_grid_dims[3]
    cdef int incremental
    # Below is for the treecode.
    cdef np.float64_t opening_angle
    # We'll store dist here so it doesn't have to be calculated twice.
    cdef np.float64_t dist
    cdef np.float64_t root_dx[3]
    cdef OctreeNode *last_node

    def __cinit__(self, np.ndarray[np.int64_t, ndim=1] top_grid_dims,
                  int nvals, int incremental = False):
        cdef int i, j, k
        self.incremental = incremental
        cdef np.int64_t pos[3]
        cdef np.float64_t *vals = <np.float64_t *> alloca(
                sizeof(np.float64_t)*nvals)
        cdef np.float64_t weight_val = 0.0
        self.nvals = nvals
        for i in range(nvals): vals[i] = 0.0

        self.top_grid_dims[0] = top_grid_dims[0]
        self.top_grid_dims[1] = top_grid_dims[1]
        self.top_grid_dims[2] = top_grid_dims[2]

        # This wouldn't be necessary if we did bitshifting...
        for i in range(80):
            self.po2[i] = 2**i
        # Cython doesn't seem to like sizeof(OctreeNode ***)
        self.root_nodes = <OctreeNode ****> \
            malloc(sizeof(void*) * top_grid_dims[0])

        # We initialize our root values to 0.0.
        for i in range(top_grid_dims[0]):
            pos[0] = i
            self.root_nodes[i] = <OctreeNode ***> \
                malloc(sizeof(OctreeNode **) * top_grid_dims[1])
            for j in range(top_grid_dims[1]):
                pos[1] = j
                self.root_nodes[i][j] = <OctreeNode **> \
                    malloc(sizeof(OctreeNode *) * top_grid_dims[1])
                for k in range(top_grid_dims[2]):
                    pos[2] = k
                    self.root_nodes[i][j][k] = OTN_initialize(
                        pos, nvals, vals, weight_val, 0, NULL)

    cdef void add_to_position(self,
                 int level, np.int64_t pos[3],
                 np.float64_t *val,
                 np.float64_t weight_val, treecode):
        cdef int i, j, k, L
        cdef OctreeNode *node
        node = self.find_on_root_level(pos, level)
        cdef np.int64_t fac
        for L in range(level):
            if self.incremental:
                OTN_add_value(node, val, weight_val, level, treecode)
            if node.children[0][0][0] == NULL:
                OTN_refine(node, self.incremental)
            # Maybe we should use bitwise operators?
            fac = self.po2[level - L - 1]
            i = (pos[0] >= fac*(2*node.pos[0]+1))
            j = (pos[1] >= fac*(2*node.pos[1]+1))
            k = (pos[2] >= fac*(2*node.pos[2]+1))
            node = node.children[i][j][k]
        OTN_add_value(node, val, weight_val, level, treecode)

    cdef OctreeNode *find_on_root_level(self, np.int64_t pos[3], int level):
        # We need this because the root level won't just have four children
        # So we find on the root level, then we traverse the tree.
        cdef np.int64_t i, j, k
        i = <np.int64_t> (pos[0] / self.po2[level])
        j = <np.int64_t> (pos[1] / self.po2[level])
        k = <np.int64_t> (pos[2] / self.po2[level])
        return self.root_nodes[i][j][k]


    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_array_to_tree(self, int level,
            np.ndarray[np.int64_t, ndim=1] pxs,
            np.ndarray[np.int64_t, ndim=1] pys,
            np.ndarray[np.int64_t, ndim=1] pzs,
            np.ndarray[np.float64_t, ndim=2] pvals,
            np.ndarray[np.float64_t, ndim=1] pweight_vals,
            int treecode = 0):
        cdef int np = pxs.shape[0]
        cdef int p
        cdef cnp.float64_t *vals
        cdef cnp.float64_t *data = <cnp.float64_t *> pvals.data
        cdef cnp.int64_t pos[3]
        for p in range(np):
            vals = data + self.nvals*p
            pos[0] = pxs[p]
            pos[1] = pys[p]
            pos[2] = pzs[p]
            self.add_to_position(level, pos, vals, pweight_vals[p], treecode)

    def add_grid_to_tree(self, int level,
                         np.ndarray[np.int64_t, ndim=1] start_index,
                         np.ndarray[np.float64_t, ndim=2] pvals,
                         np.ndarray[np.float64_t, ndim=2] wvals,
                         np.ndarray[np.int32_t, ndim=2] cm):
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_all_from_level(self, int level, int count_only = 0):
        cdef int i, j, k
        cdef int total = 0
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    total += self.count_at_level(self.root_nodes[i][j][k], level)
        if count_only: return total
        # Allocate our array
        cdef np.ndarray[np.int64_t, ndim=2] npos
        cdef np.ndarray[np.float64_t, ndim=2] nvals
        cdef np.ndarray[np.float64_t, ndim=1] nwvals
        npos = np.zeros( (total, 3), dtype='int64')
        nvals = np.zeros( (total, self.nvals), dtype='float64')
        nwvals = np.zeros( total, dtype='float64')
        cdef np.int64_t curpos = 0
        cdef np.int64_t *pdata = <np.int64_t *> npos.data
        cdef np.float64_t *vdata = <np.float64_t *> nvals.data
        cdef np.float64_t *wdata = <np.float64_t *> nwvals.data
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    curpos += self.fill_from_level(self.root_nodes[i][j][k],
                        level, curpos, pdata, vdata, wdata)
        return npos, nvals, nwvals

    cdef int count_at_level(self, OctreeNode *node, int level):
        cdef int i, j, k
        # We only really return a non-zero, calculated value if we are at the
        # level in question.
        if node.level == level:
            if self.incremental: return 1
            # We return 1 if there are no finer points at this level and zero
            # if there are
            return (node.children[0][0][0] == NULL)
        if node.children[0][0][0] == NULL: return 0
        cdef int count = 0
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    count += self.count_at_level(node.children[i][j][k], level)
        return count

    cdef int fill_from_level(self, OctreeNode *node, int level,
                              np.int64_t curpos,
                              np.int64_t *pdata,
                              np.float64_t *vdata,
                              np.float64_t *wdata):
        cdef int i, j, k
        if node.level == level:
            if node.children[0][0][0] != NULL and not self.incremental:
                return 0
            for i in range(self.nvals):
                vdata[self.nvals * curpos + i] = node.val[i]
            wdata[curpos] = node.weight_val
            pdata[curpos * 3] = node.pos[0]
            pdata[curpos * 3 + 1] = node.pos[1]
            pdata[curpos * 3 + 2] = node.pos[2]
            return 1
        if node.children[0][0][0] == NULL: return 0
        cdef np.int64_t added = 0
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    added += self.fill_from_level(node.children[i][j][k],
                            level, curpos + added, pdata, vdata, wdata)
        return added

    @cython.cdivision(True)
    cdef np.float64_t fbe_node_separation(self, OctreeNode *node1, OctreeNode *node2):
        # Find the distance between the two nodes.
        cdef np.float64_t dx1, dx2, p1, p2, dist
        cdef int i
        dist = 0.0
        for i in range(3):
            # Discover the appropriate dx for each node/dim.
            dx1 = self.root_dx[i] / (<np.float64_t> self.po2[node1.level])
            dx2 = self.root_dx[i] / (<np.float64_t> self.po2[node2.level])
            # The added term is to re-cell center the data.
            p1 = (<np.float64_t> node1.pos[i]) * dx1 + dx1/2.
            p2 = (<np.float64_t> node2.pos[i]) * dx2 + dx2/2.
            dist += (p1 - p2) * (p1 - p2)
        dist = sqrt(dist)
        return dist

    @cython.cdivision(True)
    cdef np.float64_t fbe_opening_angle(self, OctreeNode *node1,
            OctreeNode *node2):
        # Calculate the opening angle of node2 upon the center of node1.
        # In order to keep things simple, we will not assume symmetry in all
        # three directions of the octree, and we'll use the largest dimension
        # if the tree is not symmetric. This is not strictly the opening angle
        # the purest sense, but it's slightly more accurate, so it's OK.
        # This is done in code units to match the distance calculation.
        cdef np.float64_t d2, dx2, dist
        cdef np.int64_t n2
        cdef int i
        d2 = 0.0
        if node1 is node2: return 100000.0 # Just some large number.
        if self.top_grid_dims[1] == self.top_grid_dims[0] and \
                self.top_grid_dims[2] == self.top_grid_dims[0]:
            # Symmetric
            n2 = self.po2[node2.level] * self.top_grid_dims[0]
            d2 = 1. / (<np.float64_t> n2)
        else:
            # Not symmetric
            for i in range(3):
                n2 = self.po2[node2.level] * self.top_grid_dims[i]
                dx2 = 1. / (<np.float64_t> n2)
                d2 = f64max(d2, dx2)
        # Now calculate the opening angle.
        dist = self.fbe_node_separation(node1, node2)
        self.dist = dist
        return d2 / dist

    cdef void set_next(self, OctreeNode *node, int treecode):
        # This sets up the linked list, pointing node.next to the next node
        # in the iteration order.
        cdef int i, j, k
        if treecode and node.val[0] is not 0.:
            # In a treecode, we only make a new next link if this node has mass.
            self.last_node.next = node
            self.last_node = node
        elif treecode and node.val[0] is 0.:
            # No mass means it's children have no mass, no need to dig an
            # further.
            return
        else:
            # We're not doing the treecode, but we still want a linked list,
            # we don't care about val[0] necessarily.
            self.last_node.next = node
            self.last_node = node
        if node.children[0][0][0] is NULL: return
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    self.set_next(node.children[i][j][k], treecode)
        return

    cdef void set_up_next(self, OctreeNode *node):
        # This sets up a second linked list, pointing node.up_next to the next
        # node in the list that is at the same or lower (coarser) level than
        # this node. This is useful in the treecode for skipping over nodes
        # that don't need to be inspected.
        cdef OctreeNode *initial_next
        cdef OctreeNode *temp_next
        initial_next = node.next
        temp_next = node.next
        if node.next is NULL: return
        while temp_next.level > node.level:
            temp_next = temp_next.next
            if temp_next is NULL: break
        node.up_next = temp_next
        self.set_up_next(initial_next)

    def finalize(self, int treecode = 0):
        # Set up the linked list for the nodes.
        # Set treecode = 1 if nodes with no mass are to be skipped in the
        # list.
        cdef int i, j, k, sum, top_grid_total, ii, jj, kk
        self.last_node = self.root_nodes[0][0][0]
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    self.set_next(self.root_nodes[i][j][k], treecode)
        # Now we want to link to the next node in the list that is
        # on a level the same or lower (coarser) than us. This will be used
        # during a treecode search so we can skip higher-level (finer) nodes.
        sum = 1
        top_grid_total = self.top_grid_dims[0] * self.top_grid_dims[1] * \
            self.top_grid_dims[2]
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    self.set_up_next(self.root_nodes[i][j][k])
                    # Point the root_nodes.up_next to the next root_node in the
                    # list, except for the last one which stays pointing to NULL.
                    if sum < top_grid_total - 1:
                        ii = i
                        jj = j
                        kk = (k + 1) % self.top_grid_dims[2]
                        if kk < k:
                            jj = (j + 1) % self.top_grid_dims[1]
                            if jj < j:
                                ii = (i + 1) % self.top_grid_dims[0]
                        self.root_nodes[i][j][k].up_next = \
                            self.root_nodes[ii][jj][kk]
                    sum += 1

    @cython.cdivision(True)
    cdef np.float64_t fbe_main(self, np.float64_t potential, int truncate,
            np.float64_t kinetic):
        # The work is done here. Starting at the top of the linked list of
        # nodes,
        cdef np.float64_t angle, dist
        cdef OctreeNode *this_node
        cdef OctreeNode *pair_node
        this_node = self.root_nodes[0][0][0]
        while this_node is not NULL:
            # Iterate down the list to a node that either has no children and
            # is at the max_level of the tree, or to a node where
            # all of its children are massless. The second case is when data
            # from a level that isn't the deepest has been added to the tree.
            while this_node.max_level is not this_node.level:
                this_node = this_node.next
                # In case we reach the end of the list...
                if this_node is NULL: break
            if this_node is NULL: break
            if truncate and potential > kinetic:
                print('Truncating...')
                break
            pair_node = this_node.next
            while pair_node is not NULL:
                # If pair_node is massless, we can jump to up_next, because
                # nothing pair_node contains will have mass either.
                # I think that this should primarily happen for root_nodes
                # created for padding to make the region cubical.
                if pair_node.val[0] is 0.0:
                    pair_node = pair_node.up_next
                    continue
                # If pair_node is a childless node, or is a coarser node with
                # no children, we can calculate the pot
                # right now, and get a new pair_node.
                if pair_node.max_level is pair_node.level:
                    dist = self.fbe_node_separation(this_node, pair_node)
                    potential += this_node.val[0] * pair_node.val[0] / dist
                    if truncate and potential > kinetic: break
                    pair_node = pair_node.next
                    continue
                # Next, if the opening angle to pair_node is small enough,
                # calculate the potential and get a new pair_node using
                # up_next because we don't need to look at pair_node's children.
                angle = self.fbe_opening_angle(this_node, pair_node)
                if angle < self.opening_angle:
                    # self.dist was just set in fbe_opening_angle, so we
                    # can use it here without re-calculating it for these two
                    # nodes.
                    potential += this_node.val[0] * pair_node.val[0] / self.dist
                    if truncate and potential > kinetic: break
                    # We can skip all the nodes that are contained within
                    # pair_node, saving time walking the linked list.
                    pair_node = pair_node.up_next
                # If we've gotten this far, pair_node has children, but it's
                # too coarse, so we simply dig deeper using .next.
                else:
                    pair_node = pair_node.next
            # We've exhausted the pair_nodes.
            # Now we find a new this_node in the list, and do more searches
            # over pair_node.
            this_node = this_node.next
        return potential

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def find_binding_energy(self, int truncate, np.float64_t kinetic,
        np.ndarray[np.float64_t, ndim=1] root_dx, float opening_angle = 1.0):
        r"""Find the binding energy of an ensemble of data points using the
        treecode method.

        Note: The first entry of the vals array MUST be Mass. Any other
        values will be ignored, including the weight array.
        """
        # The real work is done in fbe_main(), this just sets things up
        # and returns the potential.
        cdef int i
        cdef np.float64_t potential
        potential = 0.0
        self.opening_angle = opening_angle
        for i in range(3):
            self.root_dx[i] = root_dx[i]
        potential = self.fbe_main(potential, truncate, kinetic)
        return potential

    cdef int node_ID(self, OctreeNode *node):
        # Returns an unique ID for this node based on its position and level.
        cdef int ID, i, offset, root
        cdef np.int64_t this_grid_dims[3]
        offset = 0
        root = 1
        for i in range(3):
            root *= self.top_grid_dims[i]
            this_grid_dims[i] = self.top_grid_dims[i] * 2**node.level
        for i in range(node.level):
            offset += root * 2**(3 * i)
        ID = offset + (node.pos[0] + this_grid_dims[0] * (node.pos[1] + \
            this_grid_dims[1] * node.pos[2]))
        return ID

    cdef int node_ID_on_level(self, OctreeNode *node):
        # Returns the node ID on node.level for this node.
        cdef int ID, i
        cdef np.int64_t this_grid_dims[3]
        for i in range(3):
            this_grid_dims[i] = self.top_grid_dims[i] * 2**node.level
        ID = node.pos[0] + this_grid_dims[0] * (node.pos[1] + \
            this_grid_dims[1] * node.pos[2])
        return ID

    cdef void print_node_info(self, OctreeNode *node):
        cdef int i, j, k
        line = "%d\t" % self.node_ID(node)
        if node.next is not NULL:
            line += "%d\t" % self.node_ID(node.next)
        else: line += "-1\t"
        if node.up_next is not NULL:
            line += "%d\t" % self.node_ID(node.up_next)
        else: line += "-1\t"
        line += "%d\t%d\t%d\t%d\t" % (node.level,node.pos[0],node.pos[1],node.pos[2])
        for i in range(node.nvals):
            line += "%1.5e\t" % node.val[i]
        line += "%f\t" % node.weight_val
        line += "%s\t%s\t" % (node.children[0][0][0] is not NULL, node.parent is not NULL)
        if node.children[0][0][0] is not NULL:
            nline = ""
            for i in range(2):
                for j in range(2):
                    for k in range(2):
                        nline += "%d," % self.node_ID(node.children[i][j][k])
            line += nline
        print(line)
        return

    cdef void iterate_print_nodes(self, OctreeNode *node):
        cdef int i, j, k
        self.print_node_info(node)
        if node.children[0][0][0] is NULL:
            return
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    self.iterate_print_nodes(node.children[i][j][k])
        return

    def print_all_nodes(self):
        r"""
        Prints out information about all the nodes in the octree.

        Parameters
        ----------
        None.

        Examples
        --------
        >>> octree.print_all_nodes()
        (many lines of data)
        """
        cdef int i, j, k
        sys.stdout.flush()
        sys.stderr.flush()
        line = "ID\tnext\tup_n\tlevel\tx\ty\tz\t"
        for i in range(self.nvals):
            line += "val%d\t\t" % i
        line += "weight\t\tchild?\tparent?\tchildren"
        print(line)
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    self.iterate_print_nodes(self.root_nodes[i][j][k])
        sys.stdout.flush()
        sys.stderr.flush()
        return

    def __dealloc__(self):
        cdef int i, j, k
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for k in range(self.top_grid_dims[2]):
                    OTN_free(self.root_nodes[i][j][k])
                free(self.root_nodes[i][j])
            free(self.root_nodes[i])
        free(self.root_nodes)



"""
    yt.exthook
    ~~~~~~~~~~

    Redirect imports for extensions.  This module basically makes it possible
    for us to transition from ytext.foo to yt_foo without having to
    force all extensions to upgrade at the same time.

    When a user does ``from yt.extensions.foo import bar`` it will attempt to
    import ``from yt_foo import bar`` first and when that fails it will
    try to import ``from ytext.foo import bar``.

    We're switching from namespace packages because it was just too painful for
    everybody involved.

    This is used by `yt.extensions`.

    :copyright: (c) 2015 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""
# This source code was originally in flask/exthook.py
import os
import sys


class ExtensionImporter:
    """This importer redirects imports from this submodule to other locations.
    This makes it possible to transition from the old flaskext.name to the
    newer flask_name without people having a hard time.
    """

    def __init__(self, module_choices, wrapper_module):
        self.module_choices = module_choices
        self.wrapper_module = wrapper_module
        self.prefix = wrapper_module + "."
        self.prefix_cutoff = wrapper_module.count(".") + 1

    def __eq__(self, other):
        return (
            self.__class__.__module__ == other.__class__.__module__
            and self.__class__.__name__ == other.__class__.__name__
            and self.wrapper_module == other.wrapper_module
            and self.module_choices == other.module_choices
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def install(self):
        sys.meta_path[:] = [x for x in sys.meta_path if self != x] + [self]

    def find_module(self, fullname, path=None):
        if fullname.startswith(self.prefix):
            return self

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        modname = fullname.split(".", self.prefix_cutoff)[self.prefix_cutoff]
        for path in self.module_choices:
            realname = path % modname
            try:
                __import__(realname)
            except ImportError:
                exc_type, exc_value, tb = sys.exc_info()
                # since we only establish the entry in sys.modules at the
                # very this seems to be redundant, but if recursive imports
                # happen we will call into the move import a second time.
                # On the second invocation we still don't have an entry for
                # fullname in sys.modules, but we will end up with the same
                # fake module name and that import will succeed since this
                # one already has a temporary entry in the modules dict.
                # Since this one "succeeded" temporarily that second
                # invocation now will have created a fullname entry in
                # sys.modules which we have to kill.
                sys.modules.pop(fullname, None)

                # If it's an important traceback we reraise it, otherwise
                # we swallow it and try the next choice.  The skipped frame
                # is the one from __import__ above which we don't care about
                if self.is_important_traceback(realname, tb):
                    raise exc_value.with_traceback(tb.tb_next)
                continue
            module = sys.modules[fullname] = sys.modules[realname]
            if "." not in modname:
                setattr(sys.modules[self.wrapper_module], modname, module)
            return module
        raise ImportError(f"No module named {fullname}")

    def is_important_traceback(self, important_module, tb):
        """Walks a traceback's frames and checks if any of the frames
        originated in the given important module.  If that is the case then we
        were able to import the module itself but apparently something went
        wrong when the module was imported.  (Eg: import of an import failed).
        """
        while tb is not None:
            if self.is_important_frame(important_module, tb):
                return True
            tb = tb.tb_next
        return False

    def is_important_frame(self, important_module, tb):
        """Checks a single frame if it's important."""
        g = tb.tb_frame.f_globals
        if "__name__" not in g:
            return False

        module_name = g["__name__"]

        # Python 2.7 Behavior.  Modules are cleaned up late so the
        # name shows up properly here.  Success!
        if module_name == important_module:
            return True

        # Some python versions will clean up modules so early that the
        # module name at that point is no longer set.  Try guessing from
        # the filename then.
        filename = os.path.abspath(tb.tb_frame.f_code.co_filename)
        test_string = os.path.sep + important_module.replace(".", os.path.sep)
        return (
            test_string + ".py" in filename
            or test_string + os.path.sep + "__init__.py" in filename
        )



import os

import numpy as np

from yt.config import ytcfg
from yt.fields.derived_field import DerivedField
from yt.funcs import mylog, only_on_root, parse_h5_attr
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.cosmology import Cosmology
from yt.utilities.exceptions import YTException, YTFieldNotFound
from yt.utilities.linear_interpolators import (
    BilinearFieldInterpolator,
    UnilinearFieldInterpolator,
)
from yt.utilities.on_demand_imports import _h5py as h5py

data_version = {"cloudy": 2, "apec": 3}

data_url = "http://yt-project.org/data"


def _get_data_file(table_type, data_dir=None):
    data_file = "%s_emissivity_v%d.h5" % (table_type, data_version[table_type])
    if data_dir is None:
        supp_data_dir = ytcfg.get("yt", "supp_data_dir")
        data_dir = supp_data_dir if os.path.exists(supp_data_dir) else "."
    data_path = os.path.join(data_dir, data_file)
    if not os.path.exists(data_path):
        msg = "Failed to find emissivity data file {}! Please download from {}".format(
            data_file,
            data_url,
        )
        mylog.error(msg)
        raise OSError(msg)
    return data_path


class EnergyBoundsException(YTException):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper

    def __str__(self):
        return f"Energy bounds are {self.lower:e} to {self.upper:e} keV."


class ObsoleteDataException(YTException):
    def __init__(self, table_type):
        data_file = "%s_emissivity_v%d.h5" % (table_type, data_version[table_type])
        self.msg = "X-ray emissivity data is out of date.\n"
        self.msg += f"Download the latest data from {data_url}/{data_file}."

    def __str__(self):
        return self.msg


class XrayEmissivityIntegrator:
    r"""Class for making X-ray emissivity fields. Uses hdf5 data tables
    generated from Cloudy and AtomDB/APEC.

    Initialize an XrayEmissivityIntegrator object.

    Parameters
    ----------
    table_type : string
        The type of data to use when computing the emissivity values. If "cloudy",
        a file called "cloudy_emissivity.h5" is used, for photoionized
        plasmas. If, "apec", a file called "apec_emissivity.h5" is used for
        collisionally ionized plasmas. These files contain emissivity tables
        for primordial elements and for metals at solar metallicity for the
        energy range 0.1 to 100 keV.
    redshift : float, optional
        The cosmological redshift of the source of the field. Default: 0.0.
    data_dir : string, optional
        The location to look for the data table in. If not supplied, the file
        will be looked for in the location of the YT_DEST environment variable
        or in the current working directory.
    use_metals : boolean, optional
        If set to True, the emissivity will include contributions from metals.
        Default: True
    """

    def __init__(self, table_type, redshift=0.0, data_dir=None, use_metals=True):

        filename = _get_data_file(table_type, data_dir=data_dir)
        only_on_root(mylog.info, "Loading emissivity data from %s", filename)
        in_file = h5py.File(filename, mode="r")
        if "info" in in_file.attrs:
            only_on_root(mylog.info, parse_h5_attr(in_file, "info"))
        if parse_h5_attr(in_file, "version") != data_version[table_type]:
            raise ObsoleteDataException(table_type)
        else:
            only_on_root(
                mylog.info,
                "X-ray '%s' emissivity data version: %s."
                % (table_type, parse_h5_attr(in_file, "version")),
            )

        self.log_T = in_file["log_T"][:]
        self.emissivity_primordial = in_file["emissivity_primordial"][:]
        if "log_nH" in in_file:
            self.log_nH = in_file["log_nH"][:]
        if use_metals:
            self.emissivity_metals = in_file["emissivity_metals"][:]
        self.ebin = YTArray(in_file["E"], "keV")
        in_file.close()
        self.dE = np.diff(self.ebin)
        self.emid = 0.5 * (self.ebin[1:] + self.ebin[:-1]).to("erg")
        self.redshift = redshift

    def get_interpolator(self, data_type, e_min, e_max, energy=True):
        data = getattr(self, f"emissivity_{data_type}")
        if not energy:
            data = data[..., :] / self.emid.v
        e_min = YTQuantity(e_min, "keV") * (1.0 + self.redshift)
        e_max = YTQuantity(e_max, "keV") * (1.0 + self.redshift)
        if (e_min - self.ebin[0]) / e_min < -1e-3 or (
            e_max - self.ebin[-1]
        ) / e_max > 1e-3:
            raise EnergyBoundsException(self.ebin[0], self.ebin[-1])
        e_is, e_ie = np.digitize([e_min, e_max], self.ebin)
        e_is = np.clip(e_is - 1, 0, self.ebin.size - 1)
        e_ie = np.clip(e_ie, 0, self.ebin.size - 1)

        my_dE = self.dE[e_is:e_ie].copy()
        # clip edge bins if the requested range is smaller
        my_dE[0] -= e_min - self.ebin[e_is]
        my_dE[-1] -= self.ebin[e_ie] - e_max

        interp_data = (data[..., e_is:e_ie] * my_dE).sum(axis=-1)
        if data.ndim == 2:
            emiss = UnilinearFieldInterpolator(
                np.log10(interp_data),
                [self.log_T[0], self.log_T[-1]],
                "log_T",
                truncate=True,
            )
        else:
            emiss = BilinearFieldInterpolator(
                np.log10(interp_data),
                [self.log_nH[0], self.log_nH[-1], self.log_T[0], self.log_T[-1]],
                ["log_nH", "log_T"],
                truncate=True,
            )

        return emiss


def add_xray_emissivity_field(
    ds,
    e_min,
    e_max,
    redshift=0.0,
    metallicity=("gas", "metallicity"),
    table_type="cloudy",
    data_dir=None,
    cosmology=None,
    dist=None,
    ftype="gas",
):
    r"""Create X-ray emissivity fields for a given energy range.

    Parameters
    ----------
    e_min : float
        The minimum energy in keV for the energy band.
    e_min : float
        The maximum energy in keV for the energy band.
    redshift : float, optional
        The cosmological redshift of the source of the field. Default: 0.0.
    metallicity : str or tuple of str or float, optional
        Either the name of a metallicity field or a single floating-point
        number specifying a spatially constant metallicity. Must be in
        solar units. If set to None, no metals will be assumed. Default:
        ("gas", "metallicity")
    table_type : string, optional
        The type of emissivity table to be used when creating the fields.
        Options are "cloudy" or "apec". Default: "cloudy"
    data_dir : string, optional
        The location to look for the data table in. If not supplied, the file
        will be looked for in the location of the YT_DEST environment variable
        or in the current working directory.
    cosmology : :class:`~yt.utilities.cosmology.Cosmology`, optional
        If set and redshift > 0.0, this cosmology will be used when computing the
        cosmological dependence of the emission fields. If not set, yt's default
        LCDM cosmology will be used.
    dist : (value, unit) tuple or :class:`~yt.units.yt_array.YTQuantity`, optional
        The distance to the source, used for making intensity fields. You should
        only use this if your source is nearby (not cosmological). Default: None
    ftype : string, optional
        The field type to use when creating the fields, default "gas"

    This will create at least three fields:

    "xray_emissivity_{e_min}_{e_max}_keV" (erg s^-1 cm^-3)
    "xray_luminosity_{e_min}_{e_max}_keV" (erg s^-1)
    "xray_photon_emissivity_{e_min}_{e_max}_keV" (photons s^-1 cm^-3)

    and if a redshift or distance is specified it will create two others:

    "xray_intensity_{e_min}_{e_max}_keV" (erg s^-1 cm^-3 arcsec^-2)
    "xray_photon_intensity_{e_min}_{e_max}_keV" (photons s^-1 cm^-3 arcsec^-2)

    These latter two are really only useful when making projections.

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("sloshing_nomag2_hdf5_plt_cnt_0100")
    >>> yt.add_xray_emissivity_field(ds, 0.5, 2)
    >>> p = yt.ProjectionPlot(
    ...     ds, "x", ("gas", "xray_emissivity_0.5_2_keV"), table_type="apec"
    ... )
    >>> p.save()
    """
    if not isinstance(metallicity, float) and metallicity is not None:
        try:
            metallicity = ds._get_field_info(*metallicity)
        except YTFieldNotFound as e:
            raise RuntimeError(
                f"Your dataset does not have a {metallicity} field! "
                + "Perhaps you should specify a constant metallicity instead?"
            ) from e

    if table_type == "cloudy":
        # Cloudy wants to scale by nH**2
        other_n = "H_nuclei_density"
    else:
        # APEC wants to scale by nH*ne
        other_n = "El_number_density"

    def _norm_field(field, data):
        return data[ftype, "H_nuclei_density"] * data[ftype, other_n]

    ds.add_field(
        (ftype, "norm_field"), _norm_field, units="cm**-6", sampling_type="local"
    )

    my_si = XrayEmissivityIntegrator(table_type, data_dir=data_dir, redshift=redshift)

    em_0 = my_si.get_interpolator("primordial", e_min, e_max)
    emp_0 = my_si.get_interpolator("primordial", e_min, e_max, energy=False)
    if metallicity is not None:
        em_Z = my_si.get_interpolator("metals", e_min, e_max)
        emp_Z = my_si.get_interpolator("metals", e_min, e_max, energy=False)

    def _emissivity_field(field, data):
        with np.errstate(all="ignore"):
            dd = {
                "log_nH": np.log10(data[ftype, "H_nuclei_density"]),
                "log_T": np.log10(data[ftype, "temperature"]),
            }

        my_emissivity = np.power(10, em_0(dd))
        if metallicity is not None:
            if isinstance(metallicity, DerivedField):
                my_Z = data[metallicity.name].to_value("Zsun")
            else:
                my_Z = metallicity
            my_emissivity += my_Z * np.power(10, em_Z(dd))

        my_emissivity[np.isnan(my_emissivity)] = 0

        return data[ftype, "norm_field"] * YTArray(my_emissivity, "erg*cm**3/s")

    emiss_name = (ftype, f"xray_emissivity_{e_min}_{e_max}_keV")
    ds.add_field(
        emiss_name,
        function=_emissivity_field,
        display_name=rf"\epsilon_{{X}} ({e_min}-{e_max} keV)",
        sampling_type="local",
        units="erg/cm**3/s",
    )

    def _luminosity_field(field, data):
        return data[emiss_name] * data[ftype, "mass"] / data[ftype, "density"]

    lum_name = (ftype, f"xray_luminosity_{e_min}_{e_max}_keV")
    ds.add_field(
        lum_name,
        function=_luminosity_field,
        display_name=rf"\rm{{L}}_{{X}} ({e_min}-{e_max} keV)",
        sampling_type="local",
        units="erg/s",
    )

    def _photon_emissivity_field(field, data):
        dd = {
            "log_nH": np.log10(data[ftype, "H_nuclei_density"]),
            "log_T": np.log10(data[ftype, "temperature"]),
        }

        my_emissivity = np.power(10, emp_0(dd))
        if metallicity is not None:
            if isinstance(metallicity, DerivedField):
                my_Z = data[metallicity.name].to_value("Zsun")
            else:
                my_Z = metallicity
            my_emissivity += my_Z * np.power(10, emp_Z(dd))

        return data[ftype, "norm_field"] * YTArray(my_emissivity, "photons*cm**3/s")

    phot_name = (ftype, f"xray_photon_emissivity_{e_min}_{e_max}_keV")
    ds.add_field(
        phot_name,
        function=_photon_emissivity_field,
        display_name=rf"\epsilon_{{X}} ({e_min}-{e_max} keV)",
        sampling_type="local",
        units="photons/cm**3/s",
    )

    fields = [emiss_name, lum_name, phot_name]

    if redshift > 0.0 or dist is not None:

        if dist is None:
            if cosmology is None:
                if hasattr(ds, "cosmology"):
                    cosmology = ds.cosmology
                else:
                    cosmology = Cosmology()
            D_L = cosmology.luminosity_distance(0.0, redshift)
            angular_scale = 1.0 / cosmology.angular_scale(0.0, redshift)
            dist_fac = ds.quan(
                1.0 / (4.0 * np.pi * D_L * D_L * angular_scale * angular_scale).v,
                "rad**-2",
            )
        else:
            redshift = 0.0  # Only for local sources!
            try:
                # normal behaviour, if dist is a YTQuantity
                dist = ds.quan(dist.value, dist.units)
            except AttributeError as e:
                try:
                    dist = ds.quan(*dist)
                except (RuntimeError, TypeError):
                    raise TypeError(
                        "dist should be a YTQuantity or a (value, unit) tuple!"
                    ) from e

            angular_scale = dist / ds.quan(1.0, "radian")
            dist_fac = ds.quan(
                1.0 / (4.0 * np.pi * dist * dist * angular_scale * angular_scale).v,
                "rad**-2",
            )

        ei_name = (ftype, f"xray_intensity_{e_min}_{e_max}_keV")

        def _intensity_field(field, data):
            I = dist_fac * data[emiss_name]
            return I.in_units("erg/cm**3/s/arcsec**2")

        ds.add_field(
            ei_name,
            function=_intensity_field,
            display_name=rf"I_{{X}} ({e_min}-{e_max} keV)",
            sampling_type="local",
            units="erg/cm**3/s/arcsec**2",
        )

        i_name = (ftype, f"xray_photon_intensity_{e_min}_{e_max}_keV")

        def _photon_intensity_field(field, data):
            I = (1.0 + redshift) * dist_fac * data[phot_name]
            return I.in_units("photons/cm**3/s/arcsec**2")

        ds.add_field(
            i_name,
            function=_photon_intensity_field,
            display_name=rf"I_{{X}} ({e_min}-{e_max} keV)",
            sampling_type="local",
            units="photons/cm**3/s/arcsec**2",
        )

        fields += [ei_name, i_name]

    for field in fields:
        mylog.info("Adding ('%s','%s') field.", field[0], field[1])

    return fields



from functools import partial

from yt.fields.particle_fields import sph_whitelist_fields
from yt.frontends.sph.fields import SPHFieldInfo
from yt.utilities.periodic_table import periodic_table
from yt.utilities.physical_constants import kb, mp
from yt.utilities.physical_ratios import _primordial_mass_fraction


class GadgetFieldInfo(SPHFieldInfo):
    def __init__(self, ds, field_list, slice_info=None):
        if ds.gen_hsmls:
            hsml = (("smoothing_length", ("code_length", [], None)),)
            self.known_particle_fields += hsml
        super().__init__(ds, field_list, slice_info=slice_info)

    def setup_particle_fields(self, ptype, *args, **kwargs):

        # setup some special fields that only make sense for SPH particles

        if (ptype, "FourMetalFractions") in self.ds.field_list:
            self.species_names = self._setup_four_metal_fractions(ptype)
        elif (ptype, "ElevenMetalMasses") in self.ds.field_list:
            self.species_names = self._setup_eleven_metal_masses(ptype)
        if len(self.species_names) == 0:
            self.species_names = self._check_whitelist_species_fields(ptype)

        super().setup_particle_fields(ptype, *args, **kwargs)

        if ptype in ("PartType0", "Gas"):
            self.setup_gas_particle_fields(ptype)

    def _setup_four_metal_fractions(self, ptype):
        """
        This function breaks the FourMetalFractions field (if present)
        into its four component metal fraction fields and adds
        corresponding metal density fields which will later get smoothed

        This gets used with the Gadget group0000 format
        as defined in the gadget_field_specs in frontends/gadget/definitions.py
        """
        metal_names = ["C", "O", "Si", "Fe"]

        def _Fraction(field, data, i: int):
            return data[(ptype, "FourMetalFractions")][:, i]

        def _Metal_density(field, data, i: int):
            return data[(ptype, "FourMetalFractions")][:, i] * data[(ptype, "density")]

        for i, metal_name in enumerate(metal_names):

            # add the metal fraction fields
            self.add_field(
                (ptype, metal_name + "_fraction"),
                sampling_type="particle",
                function=partial(_Fraction, i=i),
                units="",
            )

            # add the metal density fields
            self.add_field(
                (ptype, metal_name + "_density"),
                sampling_type="particle",
                function=partial(_Metal_density, i=i),
                units=self.ds.unit_system["density"],
            )

        return metal_names

    def _setup_eleven_metal_masses(self, ptype):
        """
        This function breaks the ElevenMetalMasses field (if present)
        into its eleven component metal fraction fields and adds
        corresponding metal density fields which will later get smoothed

        This gets used with the magneticum_box2_hr format
        as defined in the gadget_field_specs in frontends/gadget/definitions.py
        """
        metal_names = ["He", "C", "Ca", "O", "N", "Ne", "Mg", "S", "Si", "Fe", "Ej"]

        def _Fraction(field, data, i: int):
            return data[(ptype, "ElevenMetalMasses")][:, i] / data[(ptype, "Mass")]

        def _Metal_density(field, data, metal_name: str):
            return data[(ptype, metal_name + "_fraction")] * data[(ptype, "density")]

        for i, metal_name in enumerate(metal_names):

            # add the metal fraction fields
            self.add_field(
                (ptype, metal_name + "_fraction"),
                sampling_type="particle",
                function=partial(_Fraction, i=i),
                units="",
            )

            # add the metal density fields
            self.add_field(
                (ptype, metal_name + "_density"),
                sampling_type="particle",
                function=partial(_Metal_density, metal_name=metal_name),
                units=self.ds.unit_system["density"],
            )

        # hydrogen fraction and density
        def _h_fraction(field, data):
            ret = data[(ptype, "ElevenMetalMasses")].sum(axis=1) / data[(ptype, "Mass")]
            return 1.0 - ret

        self.add_field(
            (ptype, "H_fraction"),
            sampling_type="particle",
            function=_h_fraction,
            units="",
        )

        def _h_density(field, data):
            return data[(ptype, "H_fraction")] * data[(ptype, "density")]

        self.add_field(
            (ptype, "H_density"),
            sampling_type="particle",
            function=_h_density,
            units=self.ds.unit_system["density"],
        )

        return ["H"] + metal_names[:-1]

    def _check_whitelist_species_fields(self, ptype):
        species_names = []
        for field in self.ds.field_list:
            if (
                field[0] == ptype
                and field[1].endswith(("_fraction", "_density"))
                and field[1] in sph_whitelist_fields
            ):
                symbol, _, _ = field[1].partition("_")
                species_names.append(symbol)
        return sorted(species_names, key=lambda symbol: periodic_table[symbol].num)

    def setup_gas_particle_fields(self, ptype):
        if (ptype, "Temperature") not in self.ds.field_list:

            if (ptype, "ElectronAbundance") in self.ds.field_list:

                def _temperature(field, data):
                    # Assume cosmic abundances
                    x_H = _primordial_mass_fraction["H"]
                    gamma = 5.0 / 3.0
                    a_e = data[ptype, "ElectronAbundance"]
                    mu = 4.0 / (3.0 * x_H + 1.0 + 4.0 * x_H * a_e)
                    ret = data[ptype, "InternalEnergy"] * (gamma - 1) * mu * mp / kb
                    return ret.in_units(self.ds.unit_system["temperature"])

            else:

                def _temperature(field, data):
                    gamma = 5.0 / 3.0
                    ret = (
                        data[ptype, "InternalEnergy"]
                        * (gamma - 1)
                        * data.ds.mu
                        * mp
                        / kb
                    )
                    return ret.in_units(self.ds.unit_system["temperature"])

            self.add_field(
                (ptype, "Temperature"),
                sampling_type="particle",
                function=_temperature,
                units=self.ds.unit_system["temperature"],
            )

        self.alias((ptype, "temperature"), (ptype, "Temperature"))
        # need to do this manually since that automatic aliasing that happens
        # in the FieldInfoContainer base class has already happened at this
        # point
        self.alias(("gas", "temperature"), (ptype, "Temperature"))



from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, fake_random_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import SlicePlot


@attr(ANSWER_TEST_TAG)
def test_sliceplot_set_background_color():
    # see https://github.com/yt-project/yt/issues/3854
    ds = fake_random_ds(16)

    def create_image(filename_prefix):
        field = ("gas", "density")
        p = SlicePlot(ds, "z", field, width=1.5)
        p.set_background_color(field, color="C0")
        p.save(f"{filename_prefix}_log")
        p.set_log(("gas", "density"), False)
        p.save(f"{filename_prefix}_lin")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_sliceplot_set_background_color"
    test.answer_name = "sliceplot_set_background_color"
    yield test



import os
import shutil
import tempfile
from unittest import TestCase

import numpy as np

from yt.testing import assert_almost_equal, fake_random_ds
from yt.visualization.volume_rendering.api import (
    OpaqueSource,
    Scene,
    ZBuffer,
    create_volume_source,
)


class FakeOpaqueSource(OpaqueSource):
    # A minimal (mock) concrete implementation of OpaqueSource
    def render(self, camera, zbuffer=None):
        pass

    def _validate(self):
        pass


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class ZBufferTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        np.random.seed(0)
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_composite_vr(self):
        ds = fake_random_ds(64)
        dd = ds.sphere(ds.domain_center, 0.45 * ds.domain_width[0])
        ds.field_info[ds.field_list[0]].take_log = False

        sc = Scene()
        cam = sc.add_camera(ds)
        cam.resolution = (512, 512)
        vr = create_volume_source(dd, field=ds.field_list[0])
        vr.transfer_function.clear()
        vr.transfer_function.grey_opacity = True
        vr.transfer_function.map_to_colormap(0.0, 1.0, scale=10.0, colormap="Reds")
        sc.add_source(vr)

        cam.set_width(1.8 * ds.domain_width)
        cam.lens.setup_box_properties(cam)

        # Create Arbitrary Z-buffer
        empty = cam.lens.new_image(cam)
        z = np.empty(empty.shape[:2], dtype="float64")
        # Let's put a blue plane right through the center
        z[:] = cam.width[2] / 2.0
        empty[:, :, 2] = 1.0  # Set blue to 1's
        empty[:, :, 3] = 1.0  # Set alpha to 1's
        zbuffer = ZBuffer(empty, z)
        zsource = FakeOpaqueSource()
        zsource.set_zbuffer(zbuffer)
        sc.add_source(zsource)

        im = sc.render()
        im.write_png("composite.png")

    def test_nonrectangular_add(self):
        rgba1 = np.ones((64, 1, 4))
        z1 = np.expand_dims(np.arange(64.0), 1)

        rgba2 = np.zeros((64, 1, 4))
        z2 = np.expand_dims(np.arange(63.0, -1.0, -1.0), 1)

        exact_rgba = np.concatenate((np.ones(32), np.zeros(32)))
        exact_rgba = np.expand_dims(exact_rgba, 1)
        exact_rgba = np.dstack((exact_rgba, exact_rgba, exact_rgba, exact_rgba))

        exact_z = np.concatenate((np.arange(32.0), np.arange(31.0, -1.0, -1.0)))
        exact_z = np.expand_dims(exact_z, 1)

        buff1 = ZBuffer(rgba1, z1)
        buff2 = ZBuffer(rgba2, z2)

        buff = buff1 + buff2

        assert_almost_equal(buff.rgba, exact_rgba)
        assert_almost_equal(buff.z, exact_z)

    def test_rectangular_add(self):
        rgba1 = np.ones((8, 8, 4))
        z1 = np.arange(64.0)
        z1 = z1.reshape((8, 8))
        buff1 = ZBuffer(rgba1, z1)

        rgba2 = np.zeros((8, 8, 4))
        z2 = np.arange(63.0, -1.0, -1.0)
        z2 = z2.reshape((8, 8))
        buff2 = ZBuffer(rgba2, z2)

        buff = buff1 + buff2

        exact_rgba = np.empty((8, 8, 4), dtype=np.float64)
        exact_rgba[0:4, 0:8, :] = 1.0
        exact_rgba[4:8, 0:8, :] = 0.0

        exact_z = np.concatenate((np.arange(32.0), np.arange(31.0, -1.0, -1.0)))
        exact_z = np.expand_dims(exact_z, 1)
        exact_z = exact_z.reshape(8, 8)

        assert_almost_equal(buff.rgba, exact_rgba)
        assert_almost_equal(buff.z, exact_z)



from yt.frontends.artio.api import ARTIODataset
from yt.loaders import load
from yt.testing import (
    assert_allclose_units,
    assert_equal,
    requires_file,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    PixelizedProjectionValuesTest,
    create_obj,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
    ("deposit", "all_density"),
    ("deposit", "all_count"),
)

sizmbhloz = "sizmbhloz-clref04SNth-rs9_a0.9011/sizmbhloz-clref04SNth-rs9_a0.9011.art"


@requires_ds(sizmbhloz)
def test_sizmbhloz():
    ds = data_dir_load(sizmbhloz)
    ds.max_range = 1024 * 1024
    assert_equal(str(ds), "sizmbhloz-clref04SNth-rs9_a0.9011.art")
    dso = [None, ("sphere", ("max", (0.1, "unitary")))]
    for dobj_name in dso:
        for field in _fields:
            for axis in [0, 1, 2]:
                for weight_field in [None, ("gas", "density")]:
                    yield PixelizedProjectionValuesTest(
                        ds, axis, field, weight_field, dobj_name
                    )
            yield FieldValuesTest(ds, field, dobj_name)
        dobj = create_obj(ds, dobj_name)
        s1 = dobj[("index", "ones")].sum()
        s2 = sum(mask.sum() for block, mask in dobj.blocks)
        assert_equal(s1, s2)
    assert_equal(ds.particle_type_counts, {"N-BODY": 100000, "STAR": 110650})


@requires_file(sizmbhloz)
def test_ARTIODataset():
    assert isinstance(data_dir_load(sizmbhloz), ARTIODataset)


@requires_file(sizmbhloz)
def test_units_override():
    units_override_check(sizmbhloz)


@requires_file(sizmbhloz)
def test_particle_derived_field():
    def star_age_alias(field, data):
        # test to make sure we get back data in the correct units
        # during field detection
        return data["STAR", "age"].in_units("Myr")

    ds = load(sizmbhloz)

    ds.add_field(
        ("STAR", "new_field"),
        function=star_age_alias,
        units="Myr",
        sampling_type="particle",
    )

    ad = ds.all_data()

    assert_allclose_units(ad["STAR", "age"].in_units("Myr"), ad["STAR", "new_field"])



import numpy as np
from more_itertools import always_iterable

from yt.funcs import mylog
from yt.utilities.physical_constants import clight, hcgs, kboltz


class TransferFunction:
    r"""A transfer function governs the transmission of emission and
    absorption through a volume.

    Transfer functions are defined by boundaries, bins, and the value that
    governs transmission through that bin.  This is scaled between 0 and 1.
    When integrating through a volume the value through a given cell is
    defined by the value calculated in the transfer function.

    Parameters
    ----------
    x_bounds : tuple of floats
        The min and max for the transfer function.  Values below or above
        these values are discarded.
    nbins : int
        How many bins to calculate; in between, linear interpolation is
        used, so low values are typically fine.

    Notes
    -----
    Typically, raw transfer functions are not generated unless particular
    and specific control over the integration is desired.  Usually either
    color transfer functions, where the color values are calculated from
    color tables, or multivariate transfer functions are used.
    """

    def __init__(self, x_bounds, nbins=256):
        self.pass_through = 0
        self.nbins = nbins
        # Strip units off of x_bounds, if any
        x_bounds = [np.float64(xb) for xb in x_bounds]
        self.x_bounds = x_bounds
        self.x = np.linspace(x_bounds[0], x_bounds[1], nbins).astype("float64")
        self.y = np.zeros(nbins, dtype="float64")
        self.grad_field = -1
        self.light_source_v = self.light_source_c = np.zeros(3, "float64")
        self.features = []

    def add_gaussian(self, location, width, height):
        r"""Add a Gaussian distribution to the transfer function.

        Typically, when rendering isocontours, a Gaussian distribution is the
        easiest way to draw out features.  The spread provides a softness.
        The values are calculated as :math:`f(x) = h \exp{-(x-x_0)^2 / w}`.

        Parameters
        ----------
        location : float
            The centroid of the Gaussian (:math:`x_0` in the above equation.)
        width : float
            The relative width (:math:`w` in the above equation.)
        height : float
            The peak height (:math:`h` in the above equation.)  Note that while
            values greater 1.0 will be accepted, the values of the transmission
            function are clipped at 1.0.

        Examples
        --------

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, 1.0)
        """
        vals = height * np.exp(-((self.x - location) ** 2.0) / width)
        self.y = np.clip(np.maximum(vals, self.y), 0.0, np.inf)
        self.features.append(
            (
                "gaussian",
                f"location(x):{location:3.2g}",
                f"width(x):{width:3.2g}",
                f"height(y):{height:3.2g}",
            )
        )

    def add_line(self, start, stop):
        r"""Add a line between two points to the transmission function.

        This will accept a starting point in (x,y) and an ending point in (x,y)
        and set the values of the transmission function between those x-values
        to be along the line connecting the y values.

        Parameters
        ----------
        start : tuple of floats
            (x0, y0), the starting point.  x0 is between the bounds of the
            transfer function and y0 must be between 0.0 and 1.0.
        stop : tuple of floats
            (x1, y1), the ending point.  x1 is between the bounds of the
            transfer function and y1 must be between 0.0 and 1.0.

        Examples
        --------
        This will set the transfer function to be linear from 0.0 to 1.0,
        across the bounds of the function.

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_line((-10.0, 0.0), (-5.0, 1.0))
        """
        x0, y0 = start
        x1, y1 = stop
        slope = (y1 - y0) / (x1 - x0)
        # We create a whole new set of values and then backout the ones that do
        # not satisfy our bounding box arguments
        vals = slope * (self.x - x0) + y0
        vals[~((self.x >= x0) & (self.x <= x1))] = 0.0
        self.y = np.clip(np.maximum(vals, self.y), 0.0, np.inf)
        self.features.append(
            (
                "line",
                f"start(x,y):({start[0]:3.2g}, {start[1]:3.2g})",
                f"stop(x,y):({stop[0]:3.2g}, {stop[1]:3.2g})",
            )
        )

    def add_step(self, start, stop, value):
        r"""Adds a step function to the transfer function.

        This accepts a `start` and a `stop`, and then in between those points the
        transfer function is set to the maximum of the transfer function and
        the `value`.

        Parameters
        ----------
        start : float
            This is the beginning of the step function; must be within domain
            of the transfer function.
        stop : float
            This is the ending of the step function; must be within domain
            of the transfer function.
        value : float
            The value the transfer function will be set to between `start` and
            `stop`.  Note that the transfer function will *actually* be set to
            max(y, value) where y is the existing value of the transfer
            function.

        Examples
        --------
        Note that in this example, we have added a step function, but the
        Gaussian that already exists will "win" where it exceeds 0.5.

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-7.0, 0.01, 1.0)
        >>> tf.add_step(-8.0, -6.0, 0.5)
        """
        vals = np.zeros(self.x.shape, "float64")
        vals[(self.x >= start) & (self.x <= stop)] = value
        self.y = np.clip(np.maximum(vals, self.y), 0.0, np.inf)
        self.features.append(
            (
                "step",
                f"start(x):{start:3.2g}",
                f"stop(x):{stop:3.2g}",
                f"value(y):{value:3.2g}",
            )
        )

    def add_filtered_planck(self, wavelength, trans):
        vals = np.zeros(self.x.shape, "float64")
        nu = clight / (wavelength * 1e-8)
        nu = nu[::-1]

        for i, logT in enumerate(self.x):
            T = 10**logT
            # Black body at this nu, T
            Bnu = ((2.0 * hcgs * nu**3) / clight**2.0) / (
                np.exp(hcgs * nu / (kboltz * T)) - 1.0
            )
            # transmission
            f = Bnu * trans[::-1]
            # integrate transmission over nu
            vals[i] = np.trapz(f, nu)

        # normalize by total transmission over filter
        self.y = vals / trans.sum()  # /np.trapz(trans[::-1],nu)
        # self.y = np.clip(np.maximum(vals, self.y), 0.0, 1.0)

    def plot(self, filename):
        r"""Save an image file of the transfer function.

        This function loads up matplotlib, plots the transfer function and saves.

        Parameters
        ----------
        filename : string
            The file to save out the plot as.

        Examples
        --------

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, 1.0)
        >>> tf.plot("sample.png")
        """
        import matplotlib
        import matplotlib.pyplot as plt

        matplotlib.use("Agg")

        plt.clf()
        plt.plot(self.x, self.y, "xk-")
        plt.xlim(*self.x_bounds)
        plt.ylim(0.0, 1.0)
        plt.savefig(filename)

    def show(self):
        r"""Display an image of the transfer function

        This function loads up matplotlib and displays the current transfer function.

        Parameters
        ----------

        Examples
        --------

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, 1.0)
        >>> tf.show()
        """
        import matplotlib.pyplot as plt

        plt.clf()
        plt.plot(self.x, self.y, "xk-")
        plt.xlim(*self.x_bounds)
        plt.ylim(0.0, 1.0)
        plt.draw()

    def clear(self):
        self.y[:] = 0.0
        self.features = []

    def __repr__(self):
        disp = (
            "<Transfer Function Object>: "
            "x_bounds:(%3.2g, %3.2g) nbins:%3.2g features:%s"
            % (self.x_bounds[0], self.x_bounds[1], self.nbins, self.features)
        )
        return disp


class MultiVariateTransferFunction:
    r"""This object constructs a set of field tables that allow for
    multiple field variables to control the integration through a volume.

    The integration through a volume typically only utilizes a single field
    variable (for instance, Density) to set up and control the values
    returned at the end of the integration.  For things like isocontours,
    this is fine.  However, more complicated schema are possible by using
    this object.  For instance, density-weighted emission that produces
    colors based on the temperature of the fluid.

    Parameters
    ----------
    grey_opacity : bool
        Should opacity be calculated on a channel-by-channel basis, or
        overall?  Useful for opaque renderings. Default: False

    """

    def __init__(self, grey_opacity=False):
        self.n_field_tables = 0
        self.tables = []  # Tables are interpolation tables
        self.field_ids = [0] * 6  # This correlates fields with tables
        self.weight_field_ids = [-1] * 6  # This correlates
        self.field_table_ids = [0] * 6
        self.weight_table_ids = [-1] * 6
        self.grad_field = -1
        self.light_source_v = self.light_source_c = np.zeros(3, "float64")
        self.grey_opacity = grey_opacity

    def add_field_table(self, table, field_id, weight_field_id=-1, weight_table_id=-1):
        r"""This accepts a table describing integration.

        A "field table" is a tabulated set of values that govern the
        integration through a given field.  These are defined not only by the
        transmission coefficient, interpolated from the table itself, but the
        `field_id` that describes which of several fields the integration
        coefficient is to be calculated from.

        Parameters
        ----------
        table : `TransferFunction`
            The integration table to be added to the set of tables used during
            the integration.
        field_id : int
            Each volume has an associated set of fields.  This identifies which
            of those fields will be used to calculate the integration
            coefficient from this table.
        weight_field_id : int, optional
            If specified, the value of the field this identifies will be
            multiplied against the integration coefficient.
        weight_table_id : int, optional
            If specified, the value from the *table* this identifies will be
            multiplied against the integration coefficient.

        Notes
        -----
        This can be rather complicated.  It's recommended that if you are
        interested in manipulating this in detail that you examine the source
        code, specifically the function FIT_get_value in
        yt/_amr_utils/VolumeIntegrator.pyx.

        Examples
        --------
        This example shows how to link a new transfer function against field 0.
        Note that this by itself does not link a *channel* for integration
        against a field.  This is because the weighting system does not mandate
        that all tables contribute to a channel, only that they contribute a
        value which may be used by other field tables.

        >>> mv = MultiVariateTransferFunction()
        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-7.0, 0.01, 1.0)
        >>> mv.add_field_table(tf, 0)
        """
        self.tables.append(table)
        self.field_ids[self.n_field_tables] = field_id
        self.weight_field_ids[self.n_field_tables] = weight_field_id
        self.weight_table_ids[self.n_field_tables] = weight_table_id
        self.n_field_tables += 1

    def link_channels(self, table_id, channels=0):
        r"""Link an image channel to a field table.

        Once a field table has been added, it can be linked against a channel (any
        one of the six -- red, green, blue, red absorption, green absorption, blue
        absorption) and then the value calculated for that field table will be
        added to the integration for that channel.  Not all tables must be linked
        against channels.

        Parameters
        ----------
        table_id : int
            The 0-indexed table to link.
        channels : int or list of ints
            The channel or channels to link with this table's calculated value.


        Examples
        --------
        This example shows how to link a new transfer function against field 0, and
        then link that table against all three RGB channels.  Typically an
        absorption (or 'alpha') channel is also linked.

        >>> mv = MultiVariateTransferFunction()
        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-7.0, 0.01, 1.0)
        >>> mv.add_field_table(tf, 0)
        >>> mv.link_channels(0, [0, 1, 2])
        """
        for c in always_iterable(channels):
            self.field_table_ids[c] = table_id


class ColorTransferFunction(MultiVariateTransferFunction):
    r"""A complete set of transfer functions for standard color-mapping.

    This is the best and easiest way to set up volume rendering.  It
    creates field tables for all three colors, their alphas, and has
    support for sampling color maps and adding independent color values at
    all locations.  It will correctly set up the
    `MultiVariateTransferFunction`.

    Parameters
    ----------
    x_bounds : tuple of floats
        The min and max for the transfer function.  Values below or above
        these values are discarded.
    nbins : int
        How many bins to calculate; in between, linear interpolation is
        used, so low values are typically fine.
    grey_opacity : bool
        Should opacity be calculated on a channel-by-channel basis, or
        overall?  Useful for opaque renderings.
    """

    def __init__(self, x_bounds, nbins=256, grey_opacity=False):
        MultiVariateTransferFunction.__init__(self)
        # Strip units off of x_bounds, if any
        x_bounds = [np.float64(xb) for xb in x_bounds]
        self.x_bounds = x_bounds
        self.nbins = nbins
        # This is all compatibility and convenience.
        self.red = TransferFunction(x_bounds, nbins)
        self.green = TransferFunction(x_bounds, nbins)
        self.blue = TransferFunction(x_bounds, nbins)
        self.alpha = TransferFunction(x_bounds, nbins)
        self.funcs = (self.red, self.green, self.blue, self.alpha)
        self.grey_opacity = grey_opacity
        self.features = []

        # Now we do the multivariate stuff
        # We assign to Density, but do not weight
        for i, tf in enumerate(self.funcs[:3]):
            self.add_field_table(tf, 0, weight_table_id=3)
            self.link_channels(i, i)
        self.add_field_table(self.funcs[3], 0)
        self.link_channels(3, 3)
        # We don't have a fifth table, so the value will *always* be zero.
        # self.link_channels(4, [3,4,5])

    def add_gaussian(self, location, width, height):
        r"""Add a Gaussian distribution to the transfer function.

        Typically, when rendering isocontours, a Gaussian distribution is the
        easiest way to draw out features.  The spread provides a softness.
        The values are calculated as :math:`f(x) = h \exp{-(x-x_0)^2 / w}`.

        Parameters
        ----------
        location : float
            The centroid of the Gaussian (:math:`x_0` in the above equation.)
        width : float
            The relative width (:math:`w` in the above equation.)
        height : list of 4 float
            The peak height (:math:`h` in the above equation.)  Note that while
            values greater 1.0 will be accepted, the values of the transmission
            function are clipped at 1.0.  This must be a list, and it is in the
            order of (red, green, blue, alpha).

        Examples
        --------
        This adds a red spike.

        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, [1.0, 0.0, 0.0, 1.0])
        """
        for tf, v in zip(self.funcs, height):
            tf.add_gaussian(location, width, v)
        self.features.append(
            (
                "gaussian",
                f"location(x):{location:3.2g}",
                f"width(x):{width:3.2g}",
                "height(y):(%3.2g, %3.2g, %3.2g, %3.2g)"
                % (height[0], height[1], height[2], height[3]),
            )
        )

    def add_step(self, start, stop, value):
        r"""Adds a step function to the transfer function.

        This accepts a `start` and a `stop`, and then in between those points the
        transfer function is set to the maximum of the transfer function and
        the `value`.

        Parameters
        ----------
        start : float
            This is the beginning of the step function; must be within domain
            of the transfer function.
        stop : float
            This is the ending of the step function; must be within domain
            of the transfer function.
        value : list of 4 floats
            The value the transfer function will be set to between `start` and
            `stop`.  Note that the transfer function will *actually* be set to
            max(y, value) where y is the existing value of the transfer
            function.  This must be a list, and it is in the order of (red,
            green, blue, alpha).


        Examples
        --------
        This adds a step function that will produce a white value at > -6.0.

        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.add_step(-6.0, -5.0, [1.0, 1.0, 1.0, 1.0])
        """
        for tf, v in zip(self.funcs, value):
            tf.add_step(start, stop, v)
        self.features.append(
            (
                "step",
                f"start(x):{start:3.2g}",
                f"stop(x):{stop:3.2g}",
                "value(y):(%3.2g, %3.2g, %3.2g, %3.2g)"
                % (value[0], value[1], value[2], value[3]),
            )
        )

    def plot(self, filename):
        r"""Save an image file of the transfer function.

        This function loads up matplotlib, plots all of the constituent
        transfer functions and saves.

        Parameters
        ----------
        filename : string
            The file to save out the plot as.

        Examples
        --------

        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.add_layers(8)
        >>> tf.plot("sample.png")
        """
        from matplotlib import pyplot
        from matplotlib.ticker import FuncFormatter

        pyplot.clf()
        ax = pyplot.axes()
        i_data = np.zeros((self.alpha.x.size, self.funcs[0].y.size, 3))
        i_data[:, :, 0] = np.outer(np.ones(self.alpha.x.size), self.funcs[0].y)
        i_data[:, :, 1] = np.outer(np.ones(self.alpha.x.size), self.funcs[1].y)
        i_data[:, :, 2] = np.outer(np.ones(self.alpha.x.size), self.funcs[2].y)
        ax.imshow(i_data, origin="lower")
        ax.fill_between(
            np.arange(self.alpha.y.size),
            self.alpha.x.size * self.alpha.y,
            y2=self.alpha.x.size,
            color="white",
        )
        ax.set_xlim(0, self.alpha.x.size)
        xticks = (
            np.arange(np.ceil(self.alpha.x[0]), np.floor(self.alpha.x[-1]) + 1, 1)
            - self.alpha.x[0]
        )
        xticks *= (self.alpha.x.size - 1) / (self.alpha.x[-1] - self.alpha.x[0])
        ax.xaxis.set_ticks(xticks)

        def x_format(x, pos):
            return "%.1f" % (
                x * (self.alpha.x[-1] - self.alpha.x[0]) / (self.alpha.x.size - 1)
                + self.alpha.x[0]
            )

        ax.xaxis.set_major_formatter(FuncFormatter(x_format))
        yticks = np.linspace(0, 1, 5) * self.alpha.y.size
        ax.yaxis.set_ticks(yticks)

        def y_format(y, pos):
            return y / self.alpha.y.size

        ax.yaxis.set_major_formatter(FuncFormatter(y_format))
        ax.set_ylabel("Transmission")
        ax.set_xlabel("Value")
        pyplot.savefig(filename)

    def show(self, ax=None):
        r"""Display an image of the transfer function

        This function loads up matplotlib and displays the current transfer function.

        Parameters
        ----------

        Examples
        --------

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, 1.0)
        >>> tf.show()
        """
        from matplotlib import pyplot
        from matplotlib.ticker import FuncFormatter

        pyplot.clf()
        ax = pyplot.axes()
        i_data = np.zeros((self.alpha.x.size, self.funcs[0].y.size, 3))
        i_data[:, :, 0] = np.outer(np.ones(self.alpha.x.size), self.funcs[0].y)
        i_data[:, :, 1] = np.outer(np.ones(self.alpha.x.size), self.funcs[1].y)
        i_data[:, :, 2] = np.outer(np.ones(self.alpha.x.size), self.funcs[2].y)
        ax.imshow(i_data, origin="lower")
        ax.fill_between(
            np.arange(self.alpha.y.size),
            self.alpha.x.size * self.alpha.y,
            y2=self.alpha.x.size,
            color="white",
        )
        ax.set_xlim(0, self.alpha.x.size)
        xticks = (
            np.arange(np.ceil(self.alpha.x[0]), np.floor(self.alpha.x[-1]) + 1, 1)
            - self.alpha.x[0]
        )
        xticks *= (self.alpha.x.size - 1) / (self.alpha.x[-1] - self.alpha.x[0])
        if len(xticks) > 5:
            xticks = xticks[:: len(xticks) // 5]
        ax.xaxis.set_ticks(xticks)

        def x_format(x, pos):
            return "%.1f" % (
                x * (self.alpha.x[-1] - self.alpha.x[0]) / (self.alpha.x.size - 1)
                + self.alpha.x[0]
            )

        ax.xaxis.set_major_formatter(FuncFormatter(x_format))
        yticks = np.linspace(0, 1, 5) * self.alpha.y.size
        ax.yaxis.set_ticks(yticks)

        def y_format(y, pos):
            s = f"{y:0.2f}"
            return s

        ax.yaxis.set_major_formatter(FuncFormatter(y_format))
        ax.set_ylabel("Opacity")
        ax.set_xlabel("Value")

    def vert_cbar(
        self, resolution, log_scale, ax, label=None, label_fmt=None, *, size=10
    ):
        r"""Display an image of the transfer function

        This function loads up matplotlib and displays the current transfer function.

        Parameters
        ----------

        Examples
        --------

        >>> tf = TransferFunction((-10.0, -5.0))
        >>> tf.add_gaussian(-9.0, 0.01, 1.0)
        >>> tf.show()
        """
        from matplotlib.ticker import FuncFormatter

        if label is None:
            label = ""
        alpha = self.alpha.y
        max_alpha = alpha.max()
        i_data = np.zeros((self.alpha.x.size, self.funcs[0].y.size, 3))
        i_data[:, :, 0] = np.outer(self.funcs[0].y, np.ones(self.alpha.x.size))
        i_data[:, :, 1] = np.outer(self.funcs[1].y, np.ones(self.alpha.x.size))
        i_data[:, :, 2] = np.outer(self.funcs[2].y, np.ones(self.alpha.x.size))

        ax.imshow(i_data, origin="lower", aspect="auto")
        ax.plot(alpha, np.arange(self.alpha.y.size), "w")

        # Set TF limits based on what is visible
        visible = np.argwhere(self.alpha.y > 1.0e-3 * self.alpha.y.max())

        # Display colobar values
        xticks = (
            np.arange(np.ceil(self.alpha.x[0]), np.floor(self.alpha.x[-1]) + 1, 1)
            - self.alpha.x[0]
        )
        xticks *= (self.alpha.x.size - 1) / (self.alpha.x[-1] - self.alpha.x[0])
        if len(xticks) > 5:
            xticks = xticks[:: len(xticks) // 5]

        # Add colorbar limits to the ticks (May not give ideal results)
        xticks = np.append(visible[0], xticks)
        xticks = np.append(visible[-1], xticks)
        # remove dupes
        xticks = list(set(xticks))
        ax.yaxis.set_ticks(xticks)

        def x_format(x, pos):
            val = (
                x * (self.alpha.x[-1] - self.alpha.x[0]) / (self.alpha.x.size - 1)
                + self.alpha.x[0]
            )
            if log_scale:
                val = 10**val
            if label_fmt is None:
                if abs(val) < 1.0e-3 or abs(val) > 1.0e4:
                    if not val == 0.0:
                        e = np.floor(np.log10(abs(val)))
                        return r"${:.2f}\times 10^{{ {:d} }}$".format(
                            val / 10.0**e, int(e)
                        )
                    else:
                        return r"$0$"
                else:
                    return f"{val:.1g}"
            else:
                return label_fmt % (val)

        ax.yaxis.set_major_formatter(FuncFormatter(x_format))

        yticks = np.linspace(0, 1, 2, endpoint=True) * max_alpha
        ax.xaxis.set_ticks(yticks)

        def y_format(y, pos):
            s = f"{y:0.2f}"
            return s

        ax.xaxis.set_major_formatter(FuncFormatter(y_format))
        ax.set_xlim(0.0, max_alpha)
        ax.get_xaxis().set_ticks([])
        ax.set_ylim(visible[0], visible[-1])
        ax.tick_params(axis="y", colors="white", size=10)
        ax.set_ylabel(label, color="white", size=size * resolution / 512.0)

    def sample_colormap(self, v, w, alpha=None, colormap="gist_stern", col_bounds=None):
        r"""Add a Gaussian based on an existing colormap.

        Constructing pleasing Gaussians in a transfer function can pose some
        challenges, so this function will add a single Gaussian whose colors
        are taken from a colormap scaled between the bounds of the transfer
        function.  As with `TransferFunction.add_gaussian`, the value is
        calculated as :math:`f(x) = h \exp{-(x-x_0)^2 / w}` but with the height
        for each color calculated from the colormap.

        Parameters
        ----------
        v : float
            The value at which the Gaussian is to be added.
        w : float
            The relative width (:math:`w` in the above equation.)
        alpha : float, optional
            The alpha value height for the Gaussian
        colormap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            https://scipy-cookbook.readthedocs.io/items/Matplotlib_Show_colormaps.html .
        col_bounds: array_like, optional
            Limits ([min, max]) the values over which the colormap spans to
            these values.  Useful for sampling an entire colormap over a range
            smaller than the transfer function bounds.

        See Also
        --------
        ColorTransferFunction.add_layers : Many-at-a-time adder

        Examples
        --------

        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.sample_colormap(-7.0, 0.01, colormap="cmyt.arbre")
        """
        from yt.visualization.color_maps import _get_cmap

        v = np.float64(v)
        if col_bounds is None:
            rel = (v - self.x_bounds[0]) / (self.x_bounds[1] - self.x_bounds[0])
        else:
            rel = (v - col_bounds[0]) / (col_bounds[1] - col_bounds[0])
        cmap = _get_cmap(colormap)
        r, g, b, a = cmap(rel)
        if alpha is None:
            alpha = a
        self.add_gaussian(v, w, [r, g, b, alpha])
        mylog.debug(
            "Adding gaussian at %s with width %s and colors %s", v, w, (r, g, b, alpha)
        )

    def map_to_colormap(
        self, mi, ma, scale=1.0, colormap="gist_stern", scale_func=None
    ):
        r"""Map a range of values to a full colormap.

        Given a minimum and maximum value in the TransferFunction, map a full
        colormap over that range at an alpha level of `scale`.
        Optionally specify a scale_func function that modifies the alpha as
        a function of the transfer function value.

        Parameters
        ----------
        mi : float
            The start of the TransferFunction to map the colormap
        ma : float
            The end of the TransferFunction to map the colormap
        scale: float, optional
            The alpha value to be used for the height of the transfer function.
            Larger values will be more opaque.
        colormap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            https://scipy-cookbook.readthedocs.io/items/Matplotlib_Show_colormaps.html .
        scale_func: function(:obj:`!value`, :obj:`!minval`, :obj:`!maxval`), optional
            A user-defined function that can be used to scale the alpha channel
            as a function of the TransferFunction field values. Function maps
            value to somewhere between minval and maxval.

        Examples
        --------

        >>> def linramp(vals, minval, maxval):
        ...     return (vals - vals.min()) / (vals.max() - vals.min())
        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.map_to_colormap(-8.0, -6.0, scale=10.0, colormap="cmyt.arbre")
        >>> tf.map_to_colormap(
        ...     -6.0, -5.0, scale=10.0, colormap="cmyt.arbre", scale_func=linramp
        ... )
        """
        from yt.visualization.color_maps import _get_cmap

        mi = np.float64(mi)
        ma = np.float64(ma)
        rel0 = int(
            self.nbins * (mi - self.x_bounds[0]) / (self.x_bounds[1] - self.x_bounds[0])
        )
        rel1 = int(
            self.nbins * (ma - self.x_bounds[0]) / (self.x_bounds[1] - self.x_bounds[0])
        )
        rel0 = max(rel0, 0)
        rel1 = min(rel1, self.nbins - 1) + 1
        tomap = np.linspace(0.0, 1.0, num=rel1 - rel0)
        cmap = _get_cmap(colormap)
        cc = cmap(tomap)
        if scale_func is None:
            scale_mult = 1.0
        else:
            scale_mult = scale_func(tomap, 0.0, 1.0)
        self.red.y[rel0:rel1] = cc[:, 0] * scale_mult
        self.green.y[rel0:rel1] = cc[:, 1] * scale_mult
        self.blue.y[rel0:rel1] = cc[:, 2] * scale_mult
        self.alpha.y[rel0:rel1] = scale * cc[:, 3] * scale_mult
        self.features.append(
            (
                "map_to_colormap",
                f"start(x):{mi:3.2g}",
                f"stop(x):{ma:3.2g}",
                f"value(y):{scale:3.2g}",
            )
        )

    def add_layers(
        self,
        N,
        w=None,
        mi=None,
        ma=None,
        alpha=None,
        colormap="gist_stern",
        col_bounds=None,
    ):
        r"""Add a set of Gaussians based on an existing colormap.

        Constructing pleasing Gaussians in a transfer function can pose some
        challenges, so this function will add several evenly-spaced Gaussians
        whose colors are taken from a colormap scaled between the bounds of the
        transfer function.   For each Gaussian to be added,
        `ColorTransferFunction.sample_colormap` is called.

        Parameters
        ----------
        N : int
            How many Gaussians to add
        w : float
            The relative width of each Gaussian.  If not included, it is
            calculated as 0.001 * (max_val - min_val) / N
        mi : float, optional
            If only a subset of the data range is to have the Gaussians added,
            this is the minimum for that subset
        ma : float, optional
            If only a subset of the data range is to have the Gaussians added,
            this is the maximum for that subset
        alpha : list of floats, optional
            The alpha value height for each Gaussian.  If not supplied, it is
            set as 1.0 everywhere.
        colormap : string, optional
            An acceptable colormap.  See either yt.visualization.color_maps or
            https://scipy-cookbook.readthedocs.io/items/Matplotlib_Show_colormaps.html .
        col_bounds: array_like, optional
            Limits ([min, max]) the values over which the colormap spans to
            these values.  Useful for sampling an entire colormap over a range
            smaller than the transfer function bounds.

        See Also
        --------
        ColorTransferFunction.sample_colormap : Single Gaussian adder

        Examples
        --------

        >>> tf = ColorTransferFunction((-10.0, -5.0))
        >>> tf.add_layers(8)
        """
        if col_bounds is None:
            dist = self.x_bounds[1] - self.x_bounds[0]
            if mi is None:
                mi = self.x_bounds[0] + dist / (10.0 * N)
            if ma is None:
                ma = self.x_bounds[1] - dist / (10.0 * N)
        else:
            dist = col_bounds[1] - col_bounds[0]
            if mi is None:
                mi = col_bounds[0] + dist / (10.0 * N)
            if ma is None:
                ma = col_bounds[1] - dist / (10.0 * N)
        if w is None:
            w = 0.001 * (ma - mi) / N
            w = max(w, 1.0 / self.nbins)
        if alpha is None and self.grey_opacity:
            alpha = np.ones(N, dtype="float64")
        elif alpha is None and not self.grey_opacity:
            alpha = np.logspace(-3, 0, N)
        for v, a in zip(np.mgrid[mi : ma : N * 1j], alpha):
            self.sample_colormap(v, w, a, colormap=colormap, col_bounds=col_bounds)

    def get_colormap_image(self, height, width):
        image = np.zeros((height, width, 3), dtype="uint8")
        hvals = np.mgrid[self.x_bounds[0] : self.x_bounds[1] : height * 1j]
        for i, f in enumerate(self.funcs[:3]):
            vals = np.interp(hvals, f.x, f.y)
            image[:, :, i] = (vals[:, None] * 255).astype("uint8")
        image = image[::-1, :, :]
        return image

    def clear(self):
        for f in self.funcs:
            f.clear()
        self.features = []

    def __repr__(self):
        disp = (
            "<Color Transfer Function Object>:\n"
            + "x_bounds:[%3.2g, %3.2g] nbins:%i features:\n"
            % (self.x_bounds[0], self.x_bounds[1], self.nbins)
        )
        for f in self.features:
            disp += f"\t{str(f)}\n"
        return disp


class ProjectionTransferFunction(MultiVariateTransferFunction):
    r"""A transfer function that defines a simple projection.

    To generate an interpolated, off-axis projection through a dataset,
    this transfer function should be used.  It will create a very simple
    table that merely sums along each ray.  Note that the end product will
    need to be scaled by the total width through which the rays were cast,
    a piece of information inaccessible to the transfer function.

    Parameters
    ----------
    x_bounds : tuple of floats, optional
        If any of your values lie outside this range, they will be
        truncated.
    n_fields : int, optional
        How many fields we're going to project and pass through

    Notes
    -----
    When you use this transfer function, you may need to explicitly disable
    logging of fields.

    """

    def __init__(self, x_bounds=(-1e60, 1e60), n_fields=1):
        if n_fields > 3:
            raise NotImplementedError(
                f"supplied ${n_fields} but n_fields > 3 not implemented."
            )
        MultiVariateTransferFunction.__init__(self)
        # Strip units off of x_bounds, if any
        x_bounds = [np.float64(xb) for xb in x_bounds]
        self.x_bounds = x_bounds
        self.nbins = 2
        self.linear_mapping = TransferFunction(x_bounds, 2)
        self.linear_mapping.pass_through = 1
        self.link_channels(0, [0, 1, 2])  # same emission for all rgb, default
        for i in range(n_fields):
            self.add_field_table(self.linear_mapping, i)
            self.link_channels(i, i)
        self.link_channels(n_fields, [3, 4, 5])  # this will remove absorption


class PlanckTransferFunction(MultiVariateTransferFunction):
    """
    This sets up a planck function for multivariate emission and
    absorption.  We assume that the emission is black body, which is then
    convolved with appropriate Johnson filters for *red*, *green* and
    *blue*.  *T_bounds* and *rho_bounds* define the limits of tabulated
    emission and absorption functions.  *anorm* is a "fudge factor" that
    defines the somewhat arbitrary normalization to the scattering
    approximation: because everything is done largely unit-free, and is
    really not terribly accurate anyway, feel free to adjust this to change
    the relative amount of reddening.  Maybe in some future version this
    will be unitful.
    """

    def __init__(
        self, T_bounds, rho_bounds, nbins=256, red="R", green="V", blue="B", anorm=1e6
    ):
        MultiVariateTransferFunction.__init__(self)
        mscat = -1
        from .UBVRI import johnson_filters

        for i, f in enumerate([red, green, blue]):
            jf = johnson_filters[f]
            tf = TransferFunction(T_bounds)
            tf.add_filtered_planck(jf["wavelen"], jf["trans"])
            self.add_field_table(tf, 0, 1)
            self.link_channels(i, i)  # 0 => 0, 1 => 1, 2 => 2
            mscat = max(mscat, jf["Lchar"] ** -4)

        for i, f in enumerate([red, green, blue]):
            # Now we set up the scattering
            scat = (johnson_filters[f]["Lchar"] ** -4 / mscat) * anorm
            tf = TransferFunction(rho_bounds)
            mylog.debug("Adding: %s with relative scattering %s", f, scat)
            tf.y *= 0.0
            tf.y += scat
            self.add_field_table(tf, 1, weight_field_id=1)
            self.link_channels(i + 3, i + 3)

        self._normalize()
        self.grey_opacity = False

    def _normalize(self):
        fmax = np.array([f.y for f in self.tables[:3]])
        normal = fmax.max(axis=0)
        for f in self.tables[:3]:
            f.y = f.y / normal


if __name__ == "__main__":
    tf = ColorTransferFunction((-20, -5))
    tf.add_gaussian(-16.0, 0.4, [0.2, 0.3, 0.1])
    tf.add_gaussian(-14.0, 0.8, [0.4, 0.1, 0.2])
    tf.add_gaussian(-10.0, 1.0, [0.0, 0.0, 1.0])
    tf.plot("tf.png")



# If not otherwise specified, we are big endian
endian = ">"

fluid_fields = [
    "Density",
    "TotalEnergy",
    "XMomentumDensity",
    "YMomentumDensity",
    "ZMomentumDensity",
    "Pressure",
    "Gamma",
    "GasEnergy",
    "MetalDensitySNII",
    "MetalDensitySNIa",
    "PotentialNew",
    "PotentialOld",
]

hydro_struct = [("pad1", ">i"), ("idc", ">i"), ("iOctCh", ">i")]
for field in fluid_fields:
    hydro_struct += ((field, ">f"),)
hydro_struct += (("pad2", ">i"),)

particle_fields = [
    "particle_mass",  # stars have variable mass
    "particle_index",
    "particle_type",
    "particle_position_x",
    "particle_position_y",
    "particle_position_z",
    "particle_velocity_x",
    "particle_velocity_y",
    "particle_velocity_z",
    "particle_mass_initial",
    "particle_creation_time",
    "particle_metallicity1",
    "particle_metallicity2",
    "particle_metallicity",
]

particle_star_fields = [
    "particle_mass",
    "particle_mass_initial",
    "particle_creation_time",
    "particle_metallicity1",
    "particle_metallicity2",
    "particle_metallicity",
]


filename_pattern = {
    "amr": ["10MpcBox_", ".d"],
    "particle_header": ["PMcrd", ".DAT"],
    "particle_data": ["PMcrs", ".DAT"],
    "particle_stars": ["stars", ".dat"],
}

amr_header_struct = [
    ("jname", 1, "256s"),
    (("istep", "t", "dt", "aexpn", "ainit"), 1, "iddff"),
    (("boxh", "Om0", "Oml0", "Omb0", "hubble"), 5, "f"),
    ("nextras", 1, "i"),
    (("extra1", "extra2"), 2, "f"),
    ("lextra", 1, "512s"),
    (("min_level", "max_level"), 2, "i"),
]

particle_header_struct = [
    (
        (
            "header",
            "aexpn",
            "aexp0",
            "amplt",
            "astep",
            "istep",
            "partw",
            "tintg",
            "Ekin",
            "Ekin1",
            "Ekin2",
            "au0",
            "aeu0",
            "Nrow",
            "Ngridc",
            "Nspecies",
            "Nseed",
            "Om0",
            "Oml0",
            "hubble",
            "Wp5",
            "Ocurv",
            "Omb0",
            "extras",
            "unknown",
        ),
        1,
        "45sffffi" + "fffffff" + "iiii" + "ffffff" + "396s" + "f",
    )
]

dmparticle_header_struct = [
    (
        "header",
        "aexpn",
        "aexp0",
        "amplt",
        "astep",
        "istep",
        "partw",
        "tintg",
        "Ekin",
        "Ekin1",
        "Ekin2",
        "au0",
        "aeu0",
        "Nrow",
        "Ngridc",
        "Nspecies",
        "Nseed",
        "Om0",
        "Oml0",
        "hubble",
        "Wp5",
        "Ocurv",
        "wspecies",
        "lspecies",
        "extras",
        "boxsize",
    ),
    (1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 10, 10, 79, 1),
]

star_struct = [
    (">d", ("t_stars", "a_stars")),
    (">i", "nstars"),
    (">d", ("ws_old", "ws_oldi")),
    (">f", "particle_mass"),
    (">f", "particle_mass_initial"),
    (">f", "particle_creation_time"),
    (">f", "particle_metallicity1"),
    (">f", "particle_metallicity2"),
]

star_name_map = {
    "particle_mass": "mass",
    "particle_mass_initial": "imass",
    "particle_creation_time": "tbirth",
    "particle_metallicity1": "metallicity1",
    "particle_metallicity2": "metallicity2",
    "particle_metallicity": "metallicity",
}

constants = {
    "Y_p": 0.245,
    "gamma": 5.0 / 3.0,
    "T_CMB0": 2.726,
    "T_min": 300.0,
    "wmu": 4.0 / (8.0 - 5.0 * 0.245),
}

seek_extras = 137



import numpy as np

from yt.testing import assert_almost_equal, fake_random_ds
from yt.utilities.math_utils import euclidean_dist, periodic_dist


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_periodicity():
    # First test the simple case were we find the distance between two points
    a = [0.1, 0.1, 0.1]
    b = [0.9, 0.9, 0.9]
    period = 1.0
    dist = periodic_dist(a, b, period)
    assert_almost_equal(dist, 0.34641016151377535)
    dist = periodic_dist(a, b, period, (True, False, False))
    assert_almost_equal(dist, 1.1489125293076059)
    dist = periodic_dist(a, b, period, (False, True, False))
    assert_almost_equal(dist, 1.1489125293076059)
    dist = periodic_dist(a, b, period, (False, False, True))
    assert_almost_equal(dist, 1.1489125293076059)
    dist = periodic_dist(a, b, period, (True, True, False))
    assert_almost_equal(dist, 0.84852813742385713)
    dist = periodic_dist(a, b, period, (True, False, True))
    assert_almost_equal(dist, 0.84852813742385713)
    dist = periodic_dist(a, b, period, (False, True, True))
    assert_almost_equal(dist, 0.84852813742385713)
    dist = euclidean_dist(a, b)
    assert_almost_equal(dist, 1.3856406460551021)

    # Now test the more complicated cases where we're calculating radii based
    # on data objects
    ds = fake_random_ds(64)

    # First we test flattened data
    data = ds.all_data()
    positions = np.array([data[("index", ax)] for ax in "xyz"])
    c = [0.1, 0.1, 0.1]
    n_tup = tuple(1 for i in range(positions.ndim - 1))
    center = np.tile(
        np.reshape(np.array(c), (positions.shape[0],) + n_tup),
        (1,) + positions.shape[1:],
    )

    dist = periodic_dist(positions, center, period, ds.periodicity)
    assert_almost_equal(dist.min(), 0.00270632938683)
    assert_almost_equal(dist.max(), 0.863319074398)

    dist = euclidean_dist(positions, center)
    assert_almost_equal(dist.min(), 0.00270632938683)
    assert_almost_equal(dist.max(), 1.54531407988)

    # Then grid-like data
    data = ds.index.grids[0]
    positions = np.array([data[("index", ax)] for ax in "xyz"])
    c = [0.1, 0.1, 0.1]
    n_tup = tuple(1 for i in range(positions.ndim - 1))
    center = np.tile(
        np.reshape(np.array(c), (positions.shape[0],) + n_tup),
        (1,) + positions.shape[1:],
    )

    dist = periodic_dist(positions, center, period, ds.periodicity)
    assert_almost_equal(dist.min(), 0.00270632938683)
    assert_almost_equal(dist.max(), 0.863319074398)

    dist = euclidean_dist(positions, center)
    assert_almost_equal(dist.min(), 0.00270632938683)
    assert_almost_equal(dist.max(), 1.54531407988)



# distutils: libraries = STD_LIBS
"""
This file contains coordinate mappings between physical coordinates and those
defined on unit elements, as well as doing the corresponding intracell
interpolation on finite element data.


"""


cimport cython
cimport numpy as np
from numpy cimport ndarray

import numpy as np

from libc.math cimport fabs

from yt.utilities.lib.autogenerated_element_samplers cimport (
    Q1Function2D,
    Q1Function3D,
    Q1Jacobian2D,
    Q1Jacobian3D,
    Q2Function2D,
    Q2Jacobian2D,
    T2Function2D,
    T2Jacobian2D,
    Tet2Function3D,
    Tet2Jacobian3D,
    W1Function3D,
    W1Jacobian3D,
)


cdef extern from "platform_dep.h":
    double fmax(double x, double y) nogil

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double determinant_3x3(double* col0,
                            double* col1,
                            double* col2) nogil:
    return col0[0]*col1[1]*col2[2] - col0[0]*col1[2]*col2[1] - \
           col0[1]*col1[0]*col2[2] + col0[1]*col1[2]*col2[0] + \
           col0[2]*col1[0]*col2[1] - col0[2]*col1[1]*col2[0]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double maxnorm(double* f, int dim) nogil:
    cdef double err
    cdef int i
    err = fabs(f[0])
    for i in range(1, dim):
        err = fmax(err, fabs(f[i]))
    return err


cdef class ElementSampler:
    '''

    This is a base class for sampling the value of a finite element solution
    at an arbitrary point inside a mesh element. In general, this will be done
    by transforming the requested physical coordinate into a mapped coordinate
    system, sampling the solution in mapped coordinates, and returning the result.
    This is not to be used directly; use one of the subclasses instead.

    '''

    def __init__(self):
        self.inclusion_tol = 1.0e-8

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        pass

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_real_point(self,
                                     double* vertices,
                                     double* field_values,
                                     double* physical_x) nogil:
        cdef double val
        cdef double mapped_coord[4]

        self.map_real_to_unit(mapped_coord, vertices, physical_x)
        val = self.sample_at_unit_point(mapped_coord, field_values)
        return val


cdef class P1Sampler1D(ElementSampler):
    '''

    This implements sampling inside a linear, 1D element.

    '''

    def __init__(self):
        super(P1Sampler1D, self).__init__()
        self.num_mapped_coords = 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:
        mapped_x[0] = -1.0 + 2.0*(physical_x[0] - vertices[0]) / (vertices[1] - vertices[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0] * (1 - coord[0]) / 2.0 + vals[1] * (1.0 + coord[0]) / 2.0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol):
            return 0
        return 1


cdef class P1Sampler2D(ElementSampler):
    '''

    This implements sampling inside a linear, triangular mesh element.
    This mapping is linear and can be inverted easily. Note that this
    implementation uses triangular (or barycentric) coordinates.

    '''

    def __init__(self):
        super(P1Sampler2D, self).__init__()
        self.num_mapped_coords = 3


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:

        cdef double[3] col0
        cdef double[3] col1
        cdef double[3] col2

        col0[0] = vertices[0]
        col0[1] = vertices[1]
        col0[2] = 1.0

        col1[0] = vertices[2]
        col1[1] = vertices[3]
        col1[2] = 1.0

        col2[0] = vertices[4]
        col2[1] = vertices[5]
        col2[2] = 1.0

        det = determinant_3x3(col0, col1, col2)

        mapped_x[0] = ((vertices[3] - vertices[5])*physical_x[0] + \
                       (vertices[4] - vertices[2])*physical_x[1] + \
                       (vertices[2]*vertices[5] - vertices[4]*vertices[3])) / det

        mapped_x[1] = ((vertices[5] - vertices[1])*physical_x[0] + \
                       (vertices[0] - vertices[4])*physical_x[1] + \
                       (vertices[4]*vertices[1] - vertices[0]*vertices[5])) / det

        mapped_x[2] = 1.0 - mapped_x[1] - mapped_x[0]


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0]*coord[0] + vals[1]*coord[1] + vals[2]*coord[2]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for triangles, we check whether all mapped_coords are
        # between 0 and 1, to within the inclusion tolerance
        cdef int i
        for i in range(3):
            if (mapped_coord[i] < -self.inclusion_tol or
                mapped_coord[i] - 1.0 > self.inclusion_tol):
                return 0
        return 1


cdef class P1Sampler3D(ElementSampler):
    '''

    This implements sampling inside a linear, tetrahedral mesh element.
    This mapping is linear and can be inverted easily.

    '''


    def __init__(self):
        super(P1Sampler3D, self).__init__()
        self.num_mapped_coords = 4

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self, double* mapped_x,
                               double* vertices, double* physical_x) nogil:

        cdef int i
        cdef double d
        cdef double[3] bvec
        cdef double[3] col0
        cdef double[3] col1
        cdef double[3] col2

        # here, we express positions relative to the 4th element,
        # which is selected by vertices[9]
        for i in range(3):
            bvec[i] = physical_x[i]       - vertices[9 + i]
            col0[i] = vertices[0 + i]     - vertices[9 + i]
            col1[i] = vertices[3 + i]     - vertices[9 + i]
            col2[i] = vertices[6 + i]     - vertices[9 + i]

        d = determinant_3x3(col0, col1, col2)
        mapped_x[0] = determinant_3x3(bvec, col1, col2)/d
        mapped_x[1] = determinant_3x3(col0, bvec, col2)/d
        mapped_x[2] = determinant_3x3(col0, col1, bvec)/d
        mapped_x[3] = 1.0 - mapped_x[0] - mapped_x[1] - mapped_x[2]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self,
                                     double* coord,
                                     double* vals) nogil:
        return vals[0]*coord[0] + vals[1]*coord[1] + \
            vals[2]*coord[2] + vals[3]*coord[3]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for tetrahedra, we check whether all mapped coordinates
        # are within 0 and 1, to within the inclusion tolerance
        cdef int i
        for i in range(4):
            if (mapped_coord[i] < -self.inclusion_tol or
                mapped_coord[i] - 1.0 > self.inclusion_tol):
                return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double u, v, w
        cdef double thresh = 2.0e-2
        if mapped_coord[0] == 0:
            u = mapped_coord[1]
            v = mapped_coord[2]
            w = mapped_coord[3]
        elif mapped_coord[1] == 0:
            u = mapped_coord[2]
            v = mapped_coord[3]
            w = mapped_coord[0]
        elif mapped_coord[2] == 0:
            u = mapped_coord[1]
            v = mapped_coord[3]
            w = mapped_coord[0]
        else:
            u = mapped_coord[1]
            v = mapped_coord[2]
            w = mapped_coord[0]
        if ((u < thresh) or
            (v < thresh) or
            (w < thresh) or
            (fabs(u - 1) < thresh) or
            (fabs(v - 1) < thresh) or
            (fabs(w - 1) < thresh)):
            return 1
        return -1


cdef class NonlinearSolveSampler3D(ElementSampler):

    '''

    This is a base class for handling element samplers that require
    a nonlinear solve to invert the mapping between coordinate systems.
    To do this, we perform Newton-Raphson iteration using a specified
    system of equations with an analytic Jacobian matrix. This solver
    is hard-coded for 3D for reasons of efficiency. This is not to be
    used directly, use one of the subclasses instead.

    '''

    def __init__(self):
        super(NonlinearSolveSampler3D, self).__init__()
        self.tolerance = 1.0e-9
        self.max_iter = 10

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        '''

        A thorough description of Newton's method and modifications for global
        convergence can be found in Dennis's text "Numerical Methods for
        Unconstrained Optimization and Nonlinear Equations."

        x: solution vector; holds unit/mapped coordinates
        xk: temporary vector for holding solution of current iteration
        f: residual vector
        r, s, t: three columns of Jacobian matrix corresponding to unit/mapped
                 coordinates r, s, and t
        d: Jacobian determinant
        s_n: Newton step vector
        lam: fraction of Newton step by which to change x
        alpha: constant proportional to how much residual required to decrease.
               1e-4 is value of alpha recommended by Dennis
        err_c: Error of current iteration
        err_plus: Error of next iteration
        min_lam: minimum fraction of Newton step that the line search is allowed
                 to take. General experience suggests that lambda values smaller
                 than 1e-3 will not significantly reduce the residual, but we
                 set to 1e-6 just to be safe
        '''

        cdef int i
        cdef double d, lam
        cdef double[3] f
        cdef double[3] r
        cdef double[3] s
        cdef double[3] t
        cdef double[3] x, xk, s_n
        cdef int iterations = 0
        cdef double err_c, err_plus
        cdef double alpha = 1e-4
        cdef double min_lam = 1e-6

        # initial guess
        for i in range(3):
            x[i] = 0.0

        # initial error norm
        self.func(f, x, vertices, physical_x)
        err_c = maxnorm(f, 3)

        # begin Newton iteration
        while (err_c > self.tolerance and iterations < self.max_iter):
            self.jac(r, s, t, x, vertices, physical_x)
            d = determinant_3x3(r, s, t)

            s_n[0] = - (determinant_3x3(f, s, t)/d)
            s_n[1] = - (determinant_3x3(r, f, t)/d)
            s_n[2] = - (determinant_3x3(r, s, f)/d)
            xk[0] = x[0] + s_n[0]
            xk[1] = x[1] + s_n[1]
            xk[2] = x[2] + s_n[2]
            self.func(f, xk, vertices, physical_x)
            err_plus = maxnorm(f, 3)

            lam = 1
            while err_plus > err_c * (1. - alpha * lam) and lam > min_lam:
                lam = lam / 2
                xk[0] = x[0] + lam * s_n[0]
                xk[1] = x[1] + lam * s_n[1]
                xk[2] = x[2] + lam * s_n[2]
                self.func(f, xk, vertices, physical_x)
                err_plus = maxnorm(f, 3)

            x[0] = xk[0]
            x[1] = xk[1]
            x[2] = xk[2]
            err_c = err_plus
            iterations += 1

        if (err_c > self.tolerance):
            # we did not converge, set bogus value
            for i in range(3):
                mapped_x[i] = -99.0
        else:
            for i in range(3):
                mapped_x[i] = x[i]


cdef class Q1Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, linear, hexahedral mesh element.

    '''

    def __init__(self):
        super(Q1Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = Q1Function3D
        self.jac = Q1Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp, tm, tp

        rm = 1.0 - coord[0]
        rp = 1.0 + coord[0]
        sm = 1.0 - coord[1]
        sp = 1.0 + coord[1]
        tm = 1.0 - coord[2]
        tp = 1.0 + coord[2]

        F = vals[0]*rm*sm*tm + vals[1]*rp*sm*tm + vals[2]*rp*sp*tm + vals[3]*rm*sp*tm + \
            vals[4]*rm*sm*tp + vals[5]*rp*sm*tp + vals[6]*rp*sp*tp + vals[7]*rm*sp*tp
        return 0.125*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for hexes, the mapped coordinates all go from -1 to 1
        # if we are inside the element.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        if (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
            fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        else:
            return -1


cdef class S2Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, 20-node hexahedral mesh element.

    '''

    def __init__(self):
        super(S2Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = S2Function3D
        self.jac = S2Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, r, s, t, rm, rp, sm, sp, tm, tp

        r = coord[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = coord[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = coord[2]
        tm = 1.0 - t
        tp = 1.0 + t

        F = rm*sm*tm*(-r - s - t - 2.0)*vals[0] \
          + rp*sm*tm*( r - s - t - 2.0)*vals[1] \
          + rp*sp*tm*( r + s - t - 2.0)*vals[2] \
          + rm*sp*tm*(-r + s - t - 2.0)*vals[3] \
          + rm*sm*tp*(-r - s + t - 2.0)*vals[4] \
          + rp*sm*tp*( r - s + t - 2.0)*vals[5] \
          + rp*sp*tp*( r + s + t - 2.0)*vals[6] \
          + rm*sp*tp*(-r + s + t - 2.0)*vals[7] \
          + 2.0*(1.0 - r*r)*sm*tm*vals[8]  \
          + 2.0*rp*(1.0 - s*s)*tm*vals[9]  \
          + 2.0*(1.0 - r*r)*sp*tm*vals[10] \
          + 2.0*rm*(1.0 - s*s)*tm*vals[11] \
          + 2.0*rm*sm*(1.0 - t*t)*vals[12] \
          + 2.0*rp*sm*(1.0 - t*t)*vals[13] \
          + 2.0*rp*sp*(1.0 - t*t)*vals[14] \
          + 2.0*rm*sp*(1.0 - t*t)*vals[15] \
          + 2.0*(1.0 - r*r)*sm*tp*vals[16] \
          + 2.0*rp*(1.0 - s*s)*tp*vals[17] \
          + 2.0*(1.0 - r*r)*sp*tp*vals[18] \
          + 2.0*rm*(1.0 - s*s)*tp*vals[19]
        return 0.125*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        if (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
            fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[0]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        elif (fabs(fabs(mapped_coord[1]) - 1.0) < 1e-1 and
              fabs(fabs(mapped_coord[2]) - 1.0) < 1e-1):
            return 1
        else:
            return -1


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void S2Function3D(double* fx,
                              double* x,
                              double* vertices,
                              double* phys_x) nogil:
        cdef int i
        cdef double r, s, t, rm, rp, sm, sp, tm, tp

        r = x[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = x[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = x[2]
        tm = 1.0 - t
        tp = 1.0 + t

        for i in range(3):
            fx[i] = rm*sm*tm*(-r - s - t - 2.0)*vertices[0 + i]  \
                  + rp*sm*tm*( r - s - t - 2.0)*vertices[3 + i]  \
                  + rp*sp*tm*( r + s - t - 2.0)*vertices[6 + i]  \
                  + rm*sp*tm*(-r + s - t - 2.0)*vertices[9 + i]  \
                  + rm*sm*tp*(-r - s + t - 2.0)*vertices[12 + i] \
                  + rp*sm*tp*( r - s + t - 2.0)*vertices[15 + i] \
                  + rp*sp*tp*( r + s + t - 2.0)*vertices[18 + i] \
                  + rm*sp*tp*(-r + s + t - 2.0)*vertices[21 + i] \
                  + 2.0*(1.0 - r*r)*sm*tm*vertices[24 + i] \
                  + 2.0*rp*(1.0 - s*s)*tm*vertices[27 + i] \
                  + 2.0*(1.0 - r*r)*sp*tm*vertices[30 + i] \
                  + 2.0*rm*(1.0 - s*s)*tm*vertices[33 + i] \
                  + 2.0*rm*sm*(1.0 - t*t)*vertices[36 + i] \
                  + 2.0*rp*sm*(1.0 - t*t)*vertices[39 + i] \
                  + 2.0*rp*sp*(1.0 - t*t)*vertices[42 + i] \
                  + 2.0*rm*sp*(1.0 - t*t)*vertices[45 + i] \
                  + 2.0*(1.0 - r*r)*sm*tp*vertices[48 + i] \
                  + 2.0*rp*(1.0 - s*s)*tp*vertices[51 + i] \
                  + 2.0*(1.0 - r*r)*sp*tp*vertices[54 + i] \
                  + 2.0*rm*(1.0 - s*s)*tp*vertices[57 + i] \
                  - 8.0*phys_x[i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void S2Jacobian3D(double* rcol,
                              double* scol,
                              double* tcol,
                              double* x,
                              double* vertices,
                              double* phys_x) nogil:
        cdef int i
        cdef double r, s, t, rm, rp, sm, sp, tm, tp

        r = x[0]
        rm = 1.0 - r
        rp = 1.0 + r

        s = x[1]
        sm = 1.0 - s
        sp = 1.0 + s

        t = x[2]
        tm = 1.0 - t
        tp = 1.0 + t

        for i in range(3):
            rcol[i] = (sm*tm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (sm*tm*(r - s - t - 2.0) + rp*sm*tm)*vertices[3  + i] \
                    + (sp*tm*(r + s - t - 2.0) + rp*sp*tm)*vertices[6  + i] \
                    + (sp*tm*(r - s + t + 2.0) - rm*sp*tm)*vertices[9  + i] \
                    + (sm*tp*(r + s - t + 2.0) - rm*sm*tp)*vertices[12 + i] \
                    + (sm*tp*(r - s + t - 2.0) + rp*sm*tp)*vertices[15 + i] \
                    + (sp*tp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (sp*tp*(r - s - t + 2.0) - rm*sp*tp)*vertices[21 + i] \
                    - 4.0*r*sm*tm*vertices[24 + i] \
                    + 2.0*(1.0 - s*s)*tm*vertices[27 + i] \
                    - 4.0*r*sp*tm*vertices[30 + i] \
                    - 2.0*(1.0 - s*s)*tm*vertices[33 + i] \
                    - 2.0*sm*(1.0 - t*t)*vertices[36 + i] \
                    + 2.0*sm*(1.0 - t*t)*vertices[39 + i] \
                    + 2.0*sp*(1.0 - t*t)*vertices[42 + i] \
                    - 2.0*sp*(1.0 - t*t)*vertices[45 + i] \
                    - 4.0*r*sm*tp*vertices[48 + i] \
                    + 2.0*(1.0 - s*s)*tp*vertices[51 + i] \
                    - 4.0*r*sp*tp*vertices[54 + i] \
                    - 2.0*(1.0 - s*s)*tp*vertices[57 + i]
            scol[i] = ( rm*tm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (-rp*tm*(r - s - t - 2.0) - rp*sm*tm)*vertices[3  + i] \
                    + ( rp*tm*(r + s - t - 2.0) + rp*sp*tm)*vertices[6  + i] \
                    + (-rm*tm*(r - s + t + 2.0) + rm*sp*tm)*vertices[9  + i] \
                    + ( rm*tp*(r + s - t + 2.0) - rm*sm*tp)*vertices[12 + i] \
                    + (-rp*tp*(r - s + t - 2.0) - rp*sm*tp)*vertices[15 + i] \
                    + ( rp*tp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (-rm*tp*(r - s - t + 2.0) + rm*sp*tp)*vertices[21 + i] \
                    - 2.0*(1.0 - r*r)*tm*vertices[24 + i] \
                    - 4.0*rp*s*tm*vertices[27 + i] \
                    + 2.0*(1.0 - r*r)*tm*vertices[30 + i] \
                    - 4.0*rm*s*tm*vertices[33 + i] \
                    - 2.0*rm*(1.0 - t*t)*vertices[36 + i] \
                    - 2.0*rp*(1.0 - t*t)*vertices[39 + i] \
                    + 2.0*rp*(1.0 - t*t)*vertices[42 + i] \
                    + 2.0*rm*(1.0 - t*t)*vertices[45 + i] \
                    - 2.0*(1.0 - r*r)*tp*vertices[48 + i] \
                    - 4.0*rp*s*tp*vertices[51 + i] \
                    + 2.0*(1.0 - r*r)*tp*vertices[54 + i] \
                    - 4.0*rm*s*tp*vertices[57 + i]
            tcol[i] = ( rm*sm*(r + s + t + 2.0) - rm*sm*tm)*vertices[0  + i] \
                    + (-rp*sm*(r - s - t - 2.0) - rp*sm*tm)*vertices[3  + i] \
                    + (-rp*sp*(r + s - t - 2.0) - rp*sp*tm)*vertices[6  + i] \
                    + ( rm*sp*(r - s + t + 2.0) - rm*sp*tm)*vertices[9  + i] \
                    + (-rm*sm*(r + s - t + 2.0) + rm*sm*tp)*vertices[12 + i] \
                    + ( rp*sm*(r - s + t - 2.0) + rp*sm*tp)*vertices[15 + i] \
                    + ( rp*sp*(r + s + t - 2.0) + rp*sp*tp)*vertices[18 + i] \
                    + (-rm*sp*(r - s - t + 2.0) + rm*sp*tp)*vertices[21 + i] \
                    - 2.0*(1.0 - r*r)*sm*vertices[24 + i] \
                    - 2.0*rp*(1.0 - s*s)*vertices[27 + i] \
                    - 2.0*(1.0 - r*r)*sp*vertices[30 + i] \
                    - 2.0*rm*(1.0 - s*s)*vertices[33 + i] \
                    - 4.0*rm*sm*t*vertices[36 + i] \
                    - 4.0*rp*sm*t*vertices[39 + i] \
                    - 4.0*rp*sp*t*vertices[42 + i] \
                    - 4.0*rm*sp*t*vertices[45 + i] \
                    + 2.0*(1.0 - r*r)*sm*vertices[48 + i] \
                    + 2.0*rp*(1.0 - s*s)*vertices[51 + i] \
                    + 2.0*(1.0 - r*r)*sp*vertices[54 + i] \
                    + 2.0*rm*(1.0 - s*s)*vertices[57 + i]


cdef class W1Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, linear, wedge mesh element.

    '''

    def __init__(self):
        super(W1Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = W1Function3D
        self.jac = W1Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F

        F = vals[0]*(1.0 - coord[0] - coord[1])*(1.0 - coord[2]) + \
            vals[1]*coord[0]*(1.0 - coord[2]) + \
            vals[2]*coord[1]*(1.0 - coord[2]) + \
            vals[3]*(1.0 - coord[0] - coord[1])*(1.0 + coord[2]) + \
            vals[4]*coord[0]*(1.0 + coord[2]) + \
            vals[5]*coord[1]*(1.0 + coord[2])

        return F / 2.0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for wedges the bounds of the mapped coordinates are:
        #     0 <= mapped_coord[0] <= 1 - mapped_coord[1]
        #     0 <= mapped_coord[1]
        #    -1 <= mapped_coord[2] <= 1
        if (mapped_coord[0] < -self.inclusion_tol or
            mapped_coord[0] + mapped_coord[1] - 1.0 > self.inclusion_tol):
            return 0
        if (mapped_coord[1] < -self.inclusion_tol):
            return 0
        if (fabs(mapped_coord[2]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double r, s, t
        cdef double thresh = 5.0e-2
        r = mapped_coord[0]
        s = mapped_coord[1]
        t = mapped_coord[2]

        cdef int near_edge_r, near_edge_s, near_edge_t
        near_edge_r = (r < thresh) or (fabs(r + s - 1.0) < thresh)
        near_edge_s = (s < thresh)
        near_edge_t = fabs(fabs(mapped_coord[2]) - 1.0) < thresh

        # we use ray.instID to pass back whether the ray is near the
        # element boundary or not (used to annotate mesh lines)
        if (near_edge_r and near_edge_s):
            return 1
        elif (near_edge_r and near_edge_t):
            return 1
        elif (near_edge_s and near_edge_t):
            return 1
        else:
            return -1


cdef class NonlinearSolveSampler2D(ElementSampler):

    '''

    This is a base class for handling element samplers that require
    a nonlinear solve to invert the mapping between coordinate systems.
    To do this, we perform Newton-Raphson iteration using a specified
    system of equations with an analytic Jacobian matrix. This solver
    is hard-coded for 2D for reasons of efficiency. This is not to be
    used directly, use one of the subclasses instead.

    '''

    def __init__(self):
        super(NonlinearSolveSampler2D, self).__init__()
        self.tolerance = 1.0e-9
        self.max_iter = 10

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void map_real_to_unit(self,
                               double* mapped_x,
                               double* vertices,
                               double* physical_x) nogil:
        '''

        A thorough description of Newton's method and modifications for global
        convergence can be found in Dennis's text "Numerical Methods for
        Unconstrained Optimization and Nonlinear Equations."

        x: solution vector; holds unit/mapped coordinates
        xk: temporary vector for holding solution of current iteration
        f: residual vector
        A: Jacobian matrix (derivative of residual vector wrt x)
        d: Jacobian determinant
        s_n: Newton step vector
        lam: fraction of Newton step by which to change x
        alpha: constant proportional to how much residual required to decrease.
               1e-4 is value of alpha recommended by Dennis
        err_c: Error of current iteration
        err_plus: Error of next iteration
        min_lam: minimum fraction of Newton step that the line search is allowed
                 to take. General experience suggests that lambda values smaller
                 than 1e-3 will not significantly reduce the residual, but we
                 set to 1e-6 just to be safe
        '''

        cdef int i
        cdef double d, lam
        cdef double[2] f
        cdef double[2] x, xk, s_n
        cdef double[4] A
        cdef int iterations = 0
        cdef double err_c, err_plus
        cdef double alpha = 1e-4
        cdef double min_lam = 1e-6

        # initial guess
        for i in range(2):
            x[i] = 0.0

        # initial error norm
        self.func(f, x, vertices, physical_x)
        err_c = maxnorm(f, 2)

        # begin Newton iteration
        while (err_c > self.tolerance and iterations < self.max_iter):
            self.jac(&A[0], &A[2], x, vertices, physical_x)
            d = (A[0]*A[3] - A[1]*A[2])

            s_n[0] = -( A[3]*f[0] - A[2]*f[1]) / d
            s_n[1] = -(-A[1]*f[0] + A[0]*f[1]) / d
            xk[0] = x[0] + s_n[0]
            xk[1] = x[1] + s_n[1]
            self.func(f, xk, vertices, physical_x)
            err_plus = maxnorm(f, 2)

            lam = 1
            while err_plus > err_c * (1. - alpha * lam) and lam > min_lam:
                lam = lam / 2
                xk[0] = x[0] + lam * s_n[0]
                xk[1] = x[1] + lam * s_n[1]
                self.func(f, xk, vertices, physical_x)
                err_plus = maxnorm(f, 2)

            x[0] = xk[0]
            x[1] = xk[1]
            err_c = err_plus
            iterations += 1

        if (err_c > self.tolerance):
            # we did not converge, set bogus value
            for i in range(2):
                mapped_x[i] = -99.0
        else:
            for i in range(2):
                mapped_x[i] = x[i]


cdef class Q1Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, linear, quadrilateral mesh element.

    '''

    def __init__(self):
        super(Q1Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = Q1Function2D
        self.jac = Q1Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp

        rm = 1.0 - coord[0]
        rp = 1.0 + coord[0]
        sm = 1.0 - coord[1]
        sp = 1.0 + coord[1]

        F = vals[0]*rm*sm + vals[1]*rp*sm + vals[2]*rp*sp + vals[3]*rm*sp
        return 0.25*F

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for quads, we check whether the mapped_coord is between
        # -1 and 1 in both directions.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol):
            return 0
        return 1

cdef class Q2Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, quadratic, quadrilateral mesh element.

    '''

    def __init__(self):
        super(Q2Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = Q2Function2D
        self.jac = Q2Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double F, rm, rp, sm, sp
        cdef double[9] phi
        cdef double rv = 0

        zet = coord[0]
        eta = coord[1]
        zetm = coord[0] - 1.
        zetp = coord[0] + 1.
        etam = coord[1] - 1.
        etap = coord[1] + 1.

        phi[0] = zet * zetm * eta * etam / 4.
        phi[1] = zet * zetp * eta * etam / 4.
        phi[2] = zet * zetp * eta * etap / 4.
        phi[3] = zet * zetm * eta * etap / 4.
        phi[4] = zetp * zetm * eta * etam / -2.
        phi[5] = zet * zetp * etap * etam / -2.
        phi[6] = zetp * zetm * eta * etap / -2.
        phi[7] = zet * zetm * etap * etam / -2.
        phi[8] = zetp * zetm * etap * etam

        for i in range(9):
            rv += vals[i] * phi[i]

        return rv

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for quads, we check whether the mapped_coord is between
        # -1 and 1 in both directions.
        if (fabs(mapped_coord[0]) - 1.0 > self.inclusion_tol or
            fabs(mapped_coord[1]) - 1.0 > self.inclusion_tol):
            return 0
        return 1


cdef class T2Sampler2D(NonlinearSolveSampler2D):

    '''

    This implements sampling inside a 2D, quadratic, triangular mesh
    element. Note that this implementation uses canonical coordinates.

    '''

    def __init__(self):
        super(T2Sampler2D, self).__init__()
        self.num_mapped_coords = 2
        self.dim = 2
        self.func = T2Function2D
        self.jac = T2Jacobian2D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double phi0, phi1, phi2, phi3, phi4, phi5, c0sq, c1sq, c0c1

        c0sq = coord[0] * coord[0]
        c1sq = coord[1] * coord[1]
        c0c1 = coord[0] * coord[1]

        phi0 = 1 - 3 * coord[0] + 2 * c0sq - 3 * coord[1] + \
               2 * c1sq + 4 * c0c1
        phi1 = -coord[0] + 2 * c0sq
        phi2 = -coord[1] + 2 * c1sq
        phi3 = 4 * coord[0] - 4 * c0sq - 4 * c0c1
        phi4 = 4 * c0c1
        phi5 = 4 * coord[1] - 4 * c1sq - 4 * c0c1

        return vals[0]*phi0 + vals[1]*phi1 + vals[2]*phi2 + vals[3]*phi3 + \
               vals[4]*phi4 + vals[5]*phi5

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for canonical tris, we check whether the mapped_coords are between
        # 0 and 1.
        if (mapped_coord[0] < -self.inclusion_tol or \
            mapped_coord[1] < -self.inclusion_tol or \
            mapped_coord[0] + mapped_coord[1] - 1.0 > self.inclusion_tol):
            return 0
        return 1

cdef class Tet2Sampler3D(NonlinearSolveSampler3D):

    '''

    This implements sampling inside a 3D, quadratic, tetrahedral mesh
    element. Note that this implementation uses canonical coordinates.

    '''

    def __init__(self):
        super(Tet2Sampler3D, self).__init__()
        self.num_mapped_coords = 3
        self.dim = 3
        self.func = Tet2Function3D
        self.jac = Tet2Jacobian3D

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef double sample_at_unit_point(self, double* coord, double* vals) nogil:
        cdef double[10] phi
        cdef double coordsq[3]
        cdef int i
        cdef double return_value = 0

        for i in range(3):
            coordsq[i] = coord[i] * coord[i]

        phi[0] = 1 - 3 * coord[0] + 2 * coordsq[0] - 3 * coord[1] + \
                 2 * coordsq[1] - 3 * coord[2] + 2 * coordsq[2] + \
                 4 * coord[0] * coord[1] + 4 * coord[0] * coord[2] + \
                 4 * coord[1] * coord[2]
        phi[1] = -coord[0] + 2 * coordsq[0]
        phi[2] = -coord[1] + 2 * coordsq[1]
        phi[3] = -coord[2] + 2 * coordsq[2]
        phi[4] = 4 * coord[0] - 4 * coordsq[0] - 4 * coord[0] * coord[1] - \
                 4 * coord[0] * coord[2]
        phi[5] = 4 * coord[0] * coord[1]
        phi[6] = 4 * coord[1] - 4 * coordsq[1] - 4 * coord[0] * coord[1] - \
                 4 * coord[1] * coord[2]
        phi[7] = 4 * coord[2] - 4 * coordsq[2] - 4 * coord[2] * coord[0] - \
                 4 * coord[2] * coord[1]
        phi[8] = 4 * coord[0] * coord[2]
        phi[9] = 4 * coord[1] * coord[2]

        for i in range(10):
            return_value += phi[i] * vals[i]

        return return_value

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_inside(self, double* mapped_coord) nogil:
        # for canonical tets, we check whether the mapped_coords are between
        # 0 and 1.
        if (mapped_coord[0] < -self.inclusion_tol or \
            mapped_coord[1] < -self.inclusion_tol or \
            mapped_coord[2] < -self.inclusion_tol or \
            mapped_coord[0] + mapped_coord[1] + mapped_coord[2] - 1.0 > \
            self.inclusion_tol):
            return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int check_mesh_lines(self, double* mapped_coord) nogil:
        cdef double u, v
        cdef double thresh = 2.0e-2
        if mapped_coord[0] == 0:
            u = mapped_coord[1]
            v = mapped_coord[2]
        elif mapped_coord[1] == 0:
            u = mapped_coord[2]
            v = mapped_coord[0]
        elif mapped_coord[2] == 0:
            u = mapped_coord[1]
            v = mapped_coord[0]
        else:
            u = mapped_coord[1]
            v = mapped_coord[2]
        if ((u < thresh) or
            (v < thresh) or
            (fabs(u - 1) < thresh) or
            (fabs(v - 1) < thresh)):
            return 1
        return -1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_hex_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                     np.ndarray[np.float64_t, ndim=1] field_values,
                     np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef Q1Sampler3D sampler = Q1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_hex20_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef S2Sampler3D sampler = S2Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tetra_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_wedge_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    cdef W1Sampler3D sampler = W1Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)
    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_linear1D_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                          np.ndarray[np.float64_t, ndim=1] field_values,
                          np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler1D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tri_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                     np.ndarray[np.float64_t, ndim=1] field_values,
                     np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = P1Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_quad_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Q1Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_quad2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                       np.ndarray[np.float64_t, ndim=1] field_values,
                       np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Q2Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tri2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = T2Sampler2D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def test_tet2_sampler(np.ndarray[np.float64_t, ndim=2] vertices,
                      np.ndarray[np.float64_t, ndim=1] field_values,
                      np.ndarray[np.float64_t, ndim=1] physical_x):

    cdef double val

    sampler = Tet2Sampler3D()

    val = sampler.sample_at_real_point(<double*> vertices.data,
                                       <double*> field_values.data,
                                       <double*> physical_x.data)

    return val
